#!/usr/bin/env python

import ursgal
import importlib
import os
import sys
import csv
import os.path


class venndiagram_1_1_0(ursgal.UNode):
    """Venn Diagram uNode"""
    META_INFO = {
        'edit_version'       : 1.10,
        'name'               : 'Venndiagram',
        'version'            : '1.1.0',
        'release_date'       : None,
        'engine_type' : {
            'visualizer' : True,
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.svg'],
        'output_suffix'      : 'venndiagram',
        'include_in_git'     : True,
        'in_development'     : False,
        'utranslation_style' : 'venndiagram_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'venndiagram_1_1_0.py',
                },
            },
        },
        'citation' :
            'Kremer, L. P. M., Leufken, J., Oyunchimeg, P., Schulze, S. & '
            'Fufezan, C. (2016) Ursgal, Universal Python Module Combining '
            'Common Bottom-Up Proteomics Tools for Large-Scale Analysis. J. '
            'Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(venndiagram_1_1_0, self).__init__(*args, **kwargs)
        pass

    def _execute(self):
        '''
        Plot Venn Diagramm for a list of .csv result files (2-5)

        Arguments are set in uparams.py but passed to the engine by self.params attribute

        Returns:
            dict: results for the different areas e.g. dict['C-(A|B|D)']['results']

            Output file is written to the common_top_level_dir
        '''
        # Keyword Arguments:
        #     Input files (self.params['_input_file_dicts']): list of dictionaries created by multi_run
        #     column_names (list)     : list of column names (str) used for the comparison
        #                               columns are merged if more than one column name given
        #     header (str)            : header of the produced Venn diagram
        #     label_list (list)       : list of labels in the same order as the input files list
        #                               names of last_search_engine are used if no label_list given
        #     output_file_name (str)  : created by self.magic_name_generator if None
        #     opacity (float)

        print('[ -ENGINE- ] Plotting Venn diagram ..')
        venndiagram_main = self.import_engine_as_python_function()

        venn_params = {}
        translations = self.params['translations'][
            '_grouped_by_translated_key']

        output_file_name = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        if output_file_name == None:
            head, tail = self.determine_common_common_head_tail_part_of_names(
                input_file_dicts=input_file_dicts)
            output_file_name = head + tail

        translations['output_file']['output_file_incl_path'] = output_file_name

        for translated_key, translation_dict in translations.items():
            if translated_key in [
                'visualization_column_names',
                # 'visualization_label_list',
            ]:
                continue
            elif translated_key == 'visualization_font':
                venn_params['font'] = translation_dict[
                    'visualization_font']['font_type']
                venn_params['label font-size header'] = translation_dict[
                    'visualization_font']['font_size_header']
                venn_params['label font-size major'] = translation_dict[
                    'visualization_font']['font_size_major']
                venn_params['label font-size minor'] = translation_dict[
                    'visualization_font']['font_size_minor']
                venn_params['label font-size venn'] = translation_dict[
                    'visualization_font']['font_size_venn']
            elif translated_key == 'visualization_scaling_factors' :
                venn_params['cx'] = translation_dict[
                    'visualization_scaling_factors']['x_axis']
                venn_params['cy'] = translation_dict[
                    'visualization_scaling_factors']['y_axis']
            elif translated_key == 'visualization_size' :
                venn_params['width'] = translation_dict[
                    'visualization_size']['width']
                venn_params['height'] = translation_dict[
                    'visualization_size']['height']
            elif len(translation_dict) == 1:
                venn_params[translated_key] = list(
                    translation_dict.values())[0]
            else:
                print('The translatd key ', translated_key,
                      ' maps on more than one ukey, but no special rules have been defined')
                print(translation_dict)
                sys.exit(1)

        column_sets = {}
        default_label = [
            'label_A',
            'label_B',
            'label_C',
            'label_D',
            'label_E',
            'label_F'
        ]
        input_file_dicts = self.params['input_file_dicts']
        data = []
        for result_pos, result_file in enumerate(input_file_dicts):
            has_last_engine = result_file.get('last_engine', False)
            if has_last_engine is False:
                label_for_venn = '{0}'.format(result_pos)
            else:
                label_for_venn = '{0} ({1})'.format(
                    has_last_engine,
                    result_pos
                )
            data_field = (
                label_for_venn,
                os.path.join(
                    result_file['dir'],
                    result_file['file']
                )
            )
            data.append(data_field)

        all_are_csv = all([f[1].upper().endswith('.CSV') for f in data])
        assert all_are_csv == True, "VennDiagram input files all have to be .csv"
        assert len(data) <= 5, '''
            ERROR: input_file_list can only contain two to five result files,
            you can merge files before, if you need.
            Current number of files: {0}'''.format(len(data))

        '''
        NEWSTYLE:

        Keyword Arguments:

            data (list of dicts) style:
                {
                    'label' : <txt>,
                    'color' : <color>,
                    'data': set(),
                    'field': <A - E> Optional otherwise position in list
                }

            output_file (str): well...

        '''
        data_list = []
        lookup_dict = {}
        fieldnames_list = []
        for n, (engine, file_path) in enumerate(data):
            venn_dict = {}
            if str(n) in self.params['translations']['visualization_label_positions']:
                label = self.params['translations'][
                    'visualization_label_positions'][str(n)]
            else:
                label = engine
            if str(n) in self.params['translations']['visualization_color_positions']:
                color = self.params['translations'][
                    'visualization_color_positions'][str(n)]
            venn_dict['label'] = label
            venn_dict['color'] = label
            print('[ Reading  ] Venn set {0} / file #{1} : {0}'.format(
                n,
                file_path)
            )
            file_object = open(file_path, 'r')
            csv_input = csv.DictReader(
                filter(
                    lambda row: row[0] != '#', file_object
                )
            )
            #collect fieldnames
            fieldnames = csv_input.fieldnames
            for f_name in fieldnames:
                if f_name not in fieldnames_list:
                    fieldnames_list.append(f_name)

            venn_dict['data'] = set()
            for line_dict in csv_input:
                unique_identifier = ''
                for column_name in self.params['translations']['visualization_column_names']:
                    unique_identifier += '||{0}'.format(line_dict[column_name])
                venn_dict['data'].add(unique_identifier)
                if unique_identifier not in lookup_dict.keys():
                    lookup_dict[unique_identifier] = []
                lookup_dict[unique_identifier].append(line_dict)
            data_list.append(venn_dict)

        return_dict = venndiagram_main(
            data=data_list,
            **venn_params
        )

        #retrieve files corresponding to each set, only if the user wants them
        if self.params['translations']['extract_venndiagram_file'] == True:
            translation_dict_label = {
                'A' : '0',
                'B' : '1',
                'C' : '2',
                'D' : '3',
                'E' : '4',
            }

            #adding new columns for venn diagram output
            fieldnames_list.append('return_dict_nomenclature')
            fieldnames_list.append('actual_name')

            print('CREATING CSV FILE FROM VENN DIAGRAM ...')
            with open(output_file_name.replace('.svg','.csv'), 'w', newline='') as new_csvfile:
                writer = csv.DictWriter(new_csvfile, fieldnames=fieldnames_list)
                writer.writeheader()
                for key in return_dict.keys():
                    # ceate output file
                    output_name = ''
                    for character in key:
                        if character in translation_dict_label.keys():
                            name_by_user = self.params['translations'][
                                'visualization_label_positions'][translation_dict_label[character]]

                            assert '_[' not in name_by_user, print(
                                    'ERROR MESSAGE: your label should not contain "_["')

                            assert ']_' not in name_by_user, print(
                                    'ERROR MESSAGE: your label should not contain "]_"')

                            assert '(' not in name_by_user, print(
                                    'ERROR MESSAGE: your label should not contain "("')

                            assert ')' not in name_by_user, print(
                                    'ERROR MESSAGE: your label should not contain ")"')
                            output_name = output_name + name_by_user
                        else :
                            if character != '(' and character != ')':
                                output_name = output_name + '_['+character+']_'
                            else :
                                output_name = output_name + character

                    results = return_dict[key]['results']

                    for unique_id in results:
                        line_dict_list = lookup_dict[unique_id]

                        for line_dict in line_dict_list:
                            line_dict['return_dict_nomenclature'] = key
                            line_dict['actual_name'] = output_name
                            writer.writerow(line_dict)

        return return_dict



#!/usr/bin/env python
import re
from collections import defaultdict as ddict
import csv
from urllib.parse import unquote
import os

pattern = re.compile(
    r'Content-Type: application/x-Mascot; name="(?P<section>[a-z0-9]*)"'
)
peptide_match_pattern = re.compile(
    r'''
    q(?P<query>[0-9]*)_
    p(?P<rank>[0-9]*)=
    (?P<missed_cleavages>[-+0-9]*),
    (?P<peptide_mass>[0-9\.]*),
    (?P<delta>[-0-9\.]*),
    (?P<number_of_ions_matched>[0-9]*),
    (?P<sequence>[A-Z]*),
    (?P<peaks_used_from_ions1>[0-9]*),
    (?P<var_mod_string>[0-9]*),
    (?P<score>[.0-9]*),
    (?P<ion_series_found>[0-9]*),
    ''',
    re.VERBOSE
)


regexs_for_crazy_mgfs = {
    'cz' : re.compile(r'''
        msmsid:F(?P<spec_id>[0-9]*),
        quan:(?P<quant>[0-9]*),
        start:(?P<rt_start_in_minutes>\d*\.\d+|\d+),
        end:(?P<rt_end_in_minutes>\d*\.\d+|\d+),
    ''', re.VERBOSE)
}

PROTON = 1.007276466


class MascotDatParser(object):

    def __init__(self, mascot_dat_file):
        """Parse idents from mascot dat file.

        Args:
            mascot_dat_file (str): Path to mascot dat file.
        """
        self.mascot_dat_file = mascot_dat_file
        self.search_file = self._get_search_file()
        self.section_dict = self._parse_mascot_sections(self.mascot_dat_file)
        self.mods         = self._parse_mods()
        self.peptide_dict = self._assemble_peptide_info(self.section_dict)
        self.peptides     = self._create_final_lookup(
            self.peptide_dict,
            self.section_dict
        )
        self._iter = self._iter_idents()

    def _get_search_file(self):
        with open(self.mascot_dat_file) as fin:
            for line in fin:
                if line.startswith('FILE='):
                    file = line.split('=')[1].strip()
                    break
        return file

    def _parse_mascot_sections(self, in_file):
        sections = ddict(list)
        current_section = None

        with open(in_file) as mascot_file:
            for line in mascot_file:
                if line.strip() == '':
                    continue
                match = pattern.match(line)
                if match is not None:
                    current_section = match.group('section')
                else:
                    sections[current_section].append(line.strip())
        return sections

    def _assemble_peptide_info(self, section_dict):
        pep_info = {}
        for line in section_dict['peptides']:
            if line.strip() == '':
                continue
            if line.startswith('--'):
                continue
            k, v = line.strip().split('=')
            k_splitted = k.split('_')
            if len(k_splitted) == 2:
                if v == '-1':
                    continue
                else:
                    m = peptide_match_pattern.match(line)
                    if m is None:
                        print(line)
                        print('????')
                        break
                    mdict = m.groupdict()
                    section_key = '_q{query}_p{rank}'.format(**mdict)
                    pep_info[section_key] = mdict
            else:
                if 'terms' in k:
                    continue
                elif '_primary_nl' in k:
                    q, p, primary, nl = k.split('_')
                    pep_info['_{0}_{1}'.format(q, p)]['primary_nl'] = v
                elif '_subst' in k:
                    q, p, subst = k.split('_')
                    if pep_info['_{0}_{1}'.format(q, p)].get(
                        'substitution',
                        None
                    ) is None:
                        pep_info['_{0}_{1}'.format(q, p)]['substitution'] = []
                    pep_info['_{0}_{1}'.format(q, p)]['substitution'].append(v)
                elif '_db'in k:
                    # merged database stuff
                    continue
                else:
                    raise Exception(
                        'Do not know what to do with this line'
                        '\n{line}'.format(line=line)
                    )
        return pep_info

    def _create_final_lookup(self, pep_info, section_info):
        global_params = {}
        peptides      = {}
        for pair in section_info['parameters']:
            k_v = pair.split('=')
            if len(k_v) == 2:
                k, v = k_v
                global_params[k] = v
        for query_with_rank in pep_info:
            info_dict = pep_info[query_with_rank]
            mods = self._format_mods(info_dict)

            query_dict = {}
            query = query_with_rank.strip('_q').split('_')[0]
            query_list = self.section_dict['query{0}'.format(query)]
            for q in query_list:
                k_v = q.split('=')
                if len(k_v) == 2:
                    k, v = k_v
                    query_dict[k] = v

            if query_dict['title'].startswith('msmsid'):
                # CZ Title
                unqstring = unquote(query_dict['title'])
                m = regexs_for_crazy_mgfs['cz'].match(unqstring)
                if m is not None:
                    spec_id = int(m.group('spec_id'))
                    rtinminutes = m.group('rt_start_in_minutes')
                    query_dict['retention_time'] =  float(rtinminutes) / 60.
                    charge = query_dict['charge'].replace("+","")
                    query_dict['title'] = '{0}.{1}.{1}.{2}'.format(
                        os.path.splitext(
                            os.path.basename(self.search_file)
                        )[0],
                        spec_id,
                        charge
                    )
                    query_dict['spectrum_id'] = spec_id
                else:
                    print('Do not understand title {title}'.format(**query_dict))
                    exit(1)
                # title_dict = self._format_cz_title(query_dict['title'])
                # spec_id    = int(title_dict['msmsid'].strip('F'))
                # query_dict['spectrum_id'] = spec_id
                # query_dict['retention_time'] = float(title_dict['peak_rt'])
                # # del query_dict['peak_rt']
            else:
                # TPP style (as used by msconvert)
                title_dict = {}
                spec_id    = int(query_dict['scans'])
                query_dict['spectrum_id'] = spec_id
                query_dict['retention_time'] = query_dict['rtinseconds']
                del query_dict['rtinseconds']

            query_dict.update(info_dict)
            query_dict['mods'] = mods
            if spec_id not in peptides:
                peptides[spec_id] = []
            peptides[spec_id].append(query_dict)
        return peptides

    def _parse_mods(self):
        mods = {
            'variable' : {},
            'fixed'    : {}
        }
        for entry in self.section_dict['masses']:
            if 'delta' in entry:
                delta, delta_entry = entry.strip().split('=')
                mods['variable'][delta.replace('delta', '')] = delta_entry
            if 'fixed' in entry.lower():
                delta, delta_entry = entry.strip().split('=')
        return mods

    def _format_mods(self, pepdict):
        mods = []
        subs = pepdict.get('substitution', None)
        for pos, mod in enumerate(pepdict['var_mod_string']):
            if mod == '0':
                continue
            if subs is not None:
                for entry in subs:
                    splitted_list = entry.split(',')
                    for i in range(0, len(splitted_list), 3):
                        sub_pos, old_aa, new_aa = splitted_list[i:i + 3]
                        if int(sub_pos) == pos:
                            mods.append('Subst({mod}):{pos}'.format(
                                mod=new_aa,
                                pos=int(sub_pos)
                            ))
            var_mod_string = self.mods['variable'].get(mod, None)

            if var_mod_string is not None:
                var_mass, var_mod = var_mod_string.split(',')
                var_mass = float(var_mass.strip())
                var_mod = var_mod.split(' ')[0]
                mods.append('{mod}:{pos}'.format(
                    mod=var_mod,
                    pos=pos
                ))

            fix_mod_string = self.mods['fixed'].get(mod, None)
            if fix_mod_string is not None:
                fix_mass, fix_mod = fix_mod_string.split(',')
                fix_mass = float(fix_mass.strip())
                fix_mod  = fix_mod.split(' ')[0]
                mods.append('{mod}:{pos}'.format(
                    mod=fix_mod,
                    pos=pos
                ))
        return mods

    def _format_cz_title(self, title):
        title_info = {}
        # under linux, we have hexadecimals
        # under windows, its unicode
        #
        entry_list = title.split('%2c')
        if len(entry_list) == 1:
            raise Exception(
                'reading mgf with CZ title under windows is not'
                'supported now due to strange hex strings'
                'Title string supplied: {title}'.format(title=title)
            )
        for e in entry_list:
            k_v = e.split('%3a')
            if len(k_v) == 2:
                k, v = k_v
                if '%2e' in v:
                    v = v.replace('%2e', '.')
                title_info[k] = v
        return title_info

    def __getitem__(self, spectrum_id):
        return self.peptides[spectrum_id]

    def _iter_idents(self):
        for key in self.peptides:
            yield self.peptides[key]

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)


def main(input_file, output_file):

    ursgal_headers = [
        'Raw data location', 'Spectrum ID', 'Spectrum Title',
        'Exp m/z', 'Charge', 'Sequence',
        'proteinacc_start_stop_pre_post_;',
        'Mascot:Score', 'Modifications', 'Retention Time (s)',
        'Start', 'Stop', 'Calc m/z',
        'Is decoy', 'rank'
    ]

    parser = MascotDatParser(input_file)
    with open(output_file, 'wt') as fout:
        csv_writer = csv.DictWriter(fout, fieldnames=ursgal_headers)
        csv_writer.writeheader()
        for spec_id in parser.peptides:
            for ident in parser.peptides[spec_id]:
                charge = int(ident['charge'].strip('+'))
                title = ident['title']
                if '%2e' in title:
                    # where do these hexs come from?
                    title = title.replace('%2e', '.')
                d = {
                    'Raw data location': parser.search_file,
                    'Spectrum Title': title,
                    'Spectrum ID': int(ident['spectrum_id']),
                    'Exp m/z': (float(ident['peptide_mass']) / charge) + PROTON,
                    'Charge': charge,
                    'Sequence': ident['sequence'],
                    'Mascot:Score': ident['score'],
                    'Modifications': ';'.join(ident['mods']),
                    'Retention Time (s)': ident['retention_time'],
                    'rank': ident['rank']
                }
                csv_writer.writerow(d)



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import shutil

class sanitize_csv_1_0_0( ursgal.UNode ):
    """sanitize_csv_1_0_0 UNode

    Result files (.csv) are sanitized following defined parameters.
    That means, for each spectrum PSMs are compared and the
    best spectrum (spectra) is (are) chosen.

    The parameters have to be defined in the params. See the engine
    documentation for further information ( :meth:`.sanitize_csv_1_0_0._execute` ).
    """

    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Sanitize CSV',
        'version'            : '1.0.0',
        'release_date'       : None,
        'engine_type' : {
            'misc_engine' : True
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : 'sanitized',
        'in_development'     : False,
        # 'rejected_output_suffix': 'rejected',
        'include_in_git'     : True,
        'distributable'      : True,
        'group_psms'         : True,
        'utranslation_style' : 'sanitize_csv_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'sanitize_csv_1_0_0.py',
                },
            },
        },
        'citation' : \
            ''
    }


    def __init__(self, *args, **kwargs):
        super(sanitize_csv_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Result files (.csv) are sanitized following defined parameters.
        That means, for each spectrum PSMs are compared and the
        best spectrum (spectra) is (are) chosen

        Input file has to be a .csv

        Creates a _sanitized.csv file and returns its path.

        Note:

            If not specified, the validation_score_field and
            bigger_scores_better parameters are determined from the
            last engine. Therefore, if sanitize_csv_1_0_0 is applied to
            merged or processed result files, both parameters need to be
            specified.


        Available parameters:

            * score_diff_threshold (float): minimum score difference between
                the best PSM and the first rejected PSM of one spectrum
            * threshold_is_log10 (bool): True, if log10 scale has been used for
                score_diff_threshold.
            * accept_conflicting_psms (bool): If True, multiple PSMs for one
                spectrum can be reported if their score difference is below
                the threshold. If False, all PSMs for one spectrum are removed
                if the score difference between the best and secondbest PSM
                is not above the threshold, i.e. if there are conflicting PSMs
                with similar scores.
            * num_compared_psms (int): maximum number of PSMs (sorted by score,
                starting with the best scoring PSM) that are compared
            * remove_redundant_psms (bool): If True, redundant PSMs (e.g.
                the same identification reported by multiple engined) for the
                same spectrum are removed. An identification is defined by the
                combination of 'Sequence', 'Modifications' and 'Charge'.

        '''
        print('[ -ENGINE- ] Executing conversion ..')
        # self.time_point(tag = 'execution')
        sanitize_csv_main = self.import_engine_as_python_function()
        if self.params['output_file'].lower().endswith('.csv') is False:
            raise ValueError('Sanitize_csv only works for csv files.')

        output_file = os.path.join(
                self.params['output_dir_path'],
                self.params['output_file']
            )
        input_file  = os.path.join(
                self.params['input_dir_path'],
                self.params['input_file']
            )

        bigger_scores_better = self.params['translations']['bigger_scores_better']
        validation_score_field = self.params['translations']['validation_score_field']

        if bigger_scores_better is None or validation_score_field is None:
            last_engine = self.get_last_search_engine( history = self.stats['history'] )
            print('''
[ WARNING ] Sanitizing based on the raw "{}" scores since
[ WARNING ] "validation_score_field" and/or "bigger_scores_better"
[ WARNING ] were not manually specified in the UController.params.
            '''.format(last_engine))
            if last_engine is None or type(last_engine) == list:
                print('''
                    Could not determine last_search_engine from input file.
                    Got {0}
                    Please specify parameters validation_score_field and bigger_scores_better
                '''.format(last_engine))
                sys.exit(1)
            else:
                bigger_scores_better = self.UNODE_UPARAMS['bigger_scores_better']['uvalue_style_translation'][last_engine]
                validation_score_field = self.UNODE_UPARAMS['validation_score_field']['uvalue_style_translation'][last_engine]

        # if self.params['translations']['write_unfiltered_results'] is False:
        #     output_file_unfiltered = None
        # else:
        #     file_extension = self.META_INFO.get(
        #         'output_suffix',
        #         None
        #     )
        #     new_file_extension = self.META_INFO.get(
        #         'rejected_output_suffix',
        #         None
        #     )
        #     output_file_unfiltered = output_file.replace(
        #         file_extension,
        #         new_file_extension
        #     )
        #     shutil.copyfile(
        #         '{0}.u.json'.format(output_file),
        #         '{0}.u.json'.format(output_file_unfiltered)
        #     )

        sanitize_csv_main(
            input_file              = input_file,
            output_file             = output_file,
            grouped_psms            = self.params['grouped_psms'],
            validation_score_field  = validation_score_field,
            bigger_scores_better    = bigger_scores_better,
            log10_threshold         = self.params['translations']['threshold_is_log10'],
            score_diff_threshold    = self.params['translations']['score_diff_threshold'],
            accept_conflicting_psms = self.params['translations']['accept_conflicting_psms'],
            num_compared_psms       = self.params['translations']['num_compared_psms'],
            remove_redundant_psms   = self.params['translations']['remove_redundant_psms'],
        )

        # self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python
import ursgal
import os
from collections import defaultdict as ddict
import csv
import sys


class thermo_raw_file_parser_1_1_2(ursgal.UNode):
    """
    Unode for ThermoRawFileParser
    For further information visit
    https://github.com/compomics/ThermoRawFileParser

    Note:
        Please download ThermoRawFileParser manually from
        https://github.com/compomics/ThermoRawFileParser

    Reference:
    Hulstaert N, Sachsenberg T, Walzer M, Barsnes H, Martens L and 
    Perez-Riverol Y (2019) ThermoRawFileParser: modular, scalable and 
    cross-platform RAW file conversion. bioRxiv https://doi.org/10.1101/622852
    """

    META_INFO = {
        'edit_version': 1.00,
        'name': 'ThermoRawFileParser',
        'version': '1.1.2',
        'release_date': '2019-04-25',
        'utranslation_style': 'thermo_raw_file_parser_style_1',
        'input_extensions': ['.raw'],
        'output_extensions': ['.mzML', '.mgf',],
        'output_suffix': None,
        'create_own_folder': False,
        'in_development': False,
        'include_in_git': False,
        'distributable': False,
        'engine_type': {
            'converter': True,
        },
        'engine': {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'            : 'ThermoRawFileParser.exe',
                    'url'            : '',
                    'zip_md5'        : None,
                    'additional_exe' : [],
                },
            },
        },
        'citation':
        'Hulstaert N, Sachsenberg T, Walzer M, Barsnes H, Martens L and '
            'Perez-Riverol Y (2019) ThermoRawFileParser: modular, scalable and '
            'cross-platform RAW file conversion. bioRxiv https://doi.org/10.1101/622852'
    }

    def __init__(self, *args, **kwargs):
        super(thermo_raw_file_parser_1_1_2, self).__init__(*args, **kwargs)
        pass

    def preflight(self):
        '''
        Formatting the command line via self.params

        Returns:
            dict: self.params

        ThermoRawFileParser.exe  usage is (use -option=value for the optional arguments):
          -h, --help                 Prints out the options.
          -i, --input=VALUE          The raw file input.
          -o, --output=VALUE         The output directory.
          -f, --format=VALUE         The output format for the spectra (0 for MGF, 1
                                       for mzML, 2 for indexed mzML, 3 for Parquet, 4
                                       for MGF with profile data excluded)
          -m, --metadata=VALUE       The metadata output format (0 for JSON, 1 for TXT).
          -g, --gzip                 GZip the output file if this flag is specified (
                                       without value).
          -u, --s3_url[=VALUE]       Optional property to write directly the data into
                                       S3 Storage.
          -k, --s3_accesskeyid[=VALUE]
                                     Optional key for the S3 bucket to write the file
                                       output.
          -t, --s3_secretaccesskey[=VALUE]
                                     Optional key for the S3 bucket to write the file
                                       output.
          -n, --s3_bucketName[=VALUE]
                                     S3 bucket name
          -v, --verbose              Enable verbose logging.
          -e, --ignoreInstrumentErrors
                                     Ignore missing properties by the instrument.
        '''
        self.params['translations']['raw_input_file'] = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        # self.created_tmp_files.append(self.param_file_name)

        if sys.platform in ['win32']:
            self.params['command_list'] = []
        else:
            self.params['command_list'] = ['mono']

        self.params['command_list'].extend([
            self.exe,
            '-i={0}'.format(self.params['translations']['raw_input_file']),
            '-o={0}'.format(self.params['output_dir_path']),
            '-f={0}'.format(self.params['translations']['output_file_type']),
        ])

        for flag, value in self.params['translations']['thermo_raw_file_parser_options'].items():
            if value is not None:
                self.params['command_list'].append(
                    '{0}={1}'.format(flag, value)
                )
            else:
                self.params['command_list'].append(
                    '{0}'.format(flag)
                )

        return self.params


#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import unittest
import pprint

class ResetController(unittest.TestCase):
    def setUp(self):
        self.R = ursgal.UController(
            params = {
                'TEST_PARAMS' : 'New',
                # this will score score_ion_list ! :)
            },
            profile  = 'LTQ XL high res'
        )
        pprint.pprint( self.R.params )

    def test_params_setting( self ):
        self.R.params['TEST_PARAMS'] = '2nd'
        self.assertEqual(self.R.params['TEST_PARAMS'] , '2nd')

    def test_profile_was_set( self ):
        self.assertEqual(self.R.params['score_ion_list'] , ['a', 'b', 'y'])

    def test_profile_was_set2( self ):
        self.R.reset_controller()
        self.R.set_profile( 'LTQ XL low res' )

        self.assertEqual(self.R.params['instrument'] , 'low_res_ltq')

    def test_profile_was_set3( self ):
        self.R.reset_controller()
        self.R.set_profile( 'LTQ XL low res' )
        self.R.reset_controller()

        self.assertEqual(self.R.params['instrument'] , 'low_res_ltq')

    def test_reset_none_default_params(self):
        self.R.params['TEST_PARAMS_2nd'] = '2nd'
        self.R.reset_controller()
        self.assertEqual(self.R.params['TEST_PARAMS_2nd'] , '2nd')

    def test_reset_none_default_params_twice(self):
        self.R.params['TEST_PARAMS'] = '2nd'
        self.R.reset_controller()
        self.R.params['TEST_PARAMS'] = '3nd'
        self.R.reset_controller()
        self.assertEqual(self.R.params['TEST_PARAMS'] , '3nd')

    def test_reset_default_params(self):
        self.R.params['enzyme'] = 'trypsin'
        self.R.reset_controller()
        self.assertEqual(self.R.params['enzyme'] , 'trypsin')

    def test_reset_default_params_to_something(self):
        self.R.params['enzyme'] = 'argc'
        self.R.reset_controller()
        self.assertEqual(self.R.params['enzyme'] , 'argc')

    def test_sweep_params(self):
        '''Not going back to default'''
        for ppm_error in range(-100, 110, 1):
            with self.subTest(ppm_error = ppm_error):
                self.R.params['frag_mass_tolerance'] = ppm_error
                self.assertEqual(
                    self.R.params['frag_mass_tolerance'],
                    ppm_error
                )

    def test_reset_default_params_twice_incl_back_to_default(self):
        '''
        Setting back to default ...
        '''
        self.R.params['enzyme'] = 'argc'
        self.R.reset_controller()
        self.R.params['enzyme'] = 'trypsin'
        self.R.reset_controller()
        self.assertEqual(self.R.params['enzyme'] , 'trypsin')

    def test_reset_default_params_twice_incl_back_to_default_none(self):
        self.R.params['machine_offset_in_ppm'] = 10
        self.R.reset_controller()
        self.R.params['machine_offset_in_ppm'] = None
        self.R.reset_controller()
        self.assertEqual(
            self.R.params['machine_offset_in_ppm'],
            None
        )

    def test_reset_default_params_twice_incl_back_to_default2(self):
        self.R.params['prefix'] = 'REALLY_'
        self.R.reset_controller()
        self.R.params['prefix'] = None
        self.R.reset_controller()
        self.assertEqual(self.R.params['prefix'] , None)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""

"""
from __future__ import absolute_import
import sys
import os

import ursgal.uparams
# this is for unorthodox queries of the params.
# please use the unode functions or UParamsMapper
# to access params since they are translated,
# grouped and so on ...

from .umapmaster import UParamMapper
# from .umapmaster import UPeptideMapper

from .unode import Meta_UNode
from .unode import UNode

from .ucontroller import UController
from .ucore import COLORS

from .chemical_composition import ChemicalComposition as ChemicalComposition
from . import chemical_composition_kb
from .unimod_mapper import UnimodMapper

import ursgal.ucore
from .profiles import PROFILES

import ursgal.ukb

GlobalUnimodMapper = UnimodMapper()
base_dir = os.path.dirname( __file__)


# We store our version number in a simple text file:
version_path = os.path.join(
    os.path.dirname(__file__),
    'version.txt'
)
with open(version_path, 'r') as version_file:
    ursgal_version = version_file.read().strip()

__version__  = ursgal_version
version_info = tuple(map(int, ursgal_version.split(".")))

if not hasattr(sys, "version_info") or sys.version_info < (3, 4):
    raise RuntimeError("Ursgal requires Python 3.4 or later.")



#!/usr/bin/python
import json
import os
import ursgal
import importlib
import time
from collections import defaultdict as ddict
import csv
import operator
import sys
import hashlib
from functools import partial
import subprocess
import re
import pprint
import gzip
import copy
from ursgal import ukb


class Meta_UNode(type):
    """Metaclass for our UNode

    additional attributes
    .exe
    .translations
    .default_params
    are read and set.
    """
    XX_meta_collected_nodes = {}

    _collected_initialized_unodes = {}

    _uparam_mapper = ursgal.UParamMapper()
    # _upeptide_mapper = ursgal.UPeptideMapper()

    def __new__(cls, cls_name, cls_bases, cls_dict ):
        new_class = super(
            Meta_UNode,
            cls
        ).__new__(
            cls,
            cls_name,
            cls_bases,
            cls_dict
        )
        return new_class

    def __init__( cls, cls_name, cls_bases, cls_dict ):
        super(Meta_UNode, cls).__init__( cls_name, cls_bases, cls_dict )

    def __call__(cls, *args, **kwargs):
        initd_klass = super(Meta_UNode, cls).__call__( cls, *args, **kwargs )

        engine_path = kwargs.get('engine_path', None)
        if engine_path is None:
            # This happens only when we call our Controller
            engine = 'ucontroller'
            kwargs['engine_path'] = __file__
        else:
            all_parts = os.path.abspath( kwargs['engine_path'] ).split(os.sep)
            engine = all_parts[ -2 ]

        initd_klass.DEFAULT_PARAMS = {}
        initd_klass.UNODE_UPARAMS = {}
        initd_klass.UNODE_UPARAMS_GROUPED_BY_TAG = ddict(list)
        # initd_klass.TRANSLATIONS_GROUPED_BY_TRANSLATED_KEY = {}
        initd_klass.PARAMS_TRIGGERING_RERUN = set()
        for mDict in Meta_UNode._uparam_mapper.mapping_dicts( engine ):
            for tag in mDict['utag']:
                initd_klass.UNODE_UPARAMS_GROUPED_BY_TAG[ tag ].append(
                    mDict['ukey']
                )

            initd_klass.DEFAULT_PARAMS[ mDict['ukey'] ] = \
                mDict['default_value_translated']

            initd_klass.UNODE_UPARAMS[ mDict['ukey'] ] = mDict

            if mDict['triggers_rerun']:
                initd_klass.PARAMS_TRIGGERING_RERUN.add( mDict['ukey'] )

        translation_style = initd_klass.META_INFO.get(
            'utranslation_style',
            None
        )
        if translation_style is not None:
            Meta_UNode._uparam_mapper.lookup[ 'style_2_engine' ][ translation_style ].add(
                engine
            )

        alternative_exe_folder = initd_klass.META_INFO.get(
            'uses_unode',
            None
        )
        if alternative_exe_folder is not None:
            kwargs['engine_path'] = kwargs['engine_path'].replace(
                engine,
                alternative_exe_folder
            )
        initd_klass.exe = kwargs['engine_path']

        obligatory_methods = [
            'preflight',
            'postflight',
        ]

        for method in obligatory_methods:
            engine_method  = getattr( initd_klass, method, None)
            assert callable(engine_method), '''
            {0} class requires {1} method to be defined
            '''.format( engine, method )
        initd_klass.engine = engine

        Meta_UNode._collected_initialized_unodes[ engine ] = initd_klass
        initd_klass.meta_unodes = Meta_UNode._collected_initialized_unodes
        initd_klass.uparam_mapper = Meta_UNode._uparam_mapper
        # initd_klass.upeptide_mapper = Meta_UNode._upeptide_mapper

        if hasattr( initd_klass, '_run_after_meta_init'):
            initd_klass._after_init_meta_callback( *args, **kwargs )

        return initd_klass


class UNode(object, metaclass=Meta_UNode):
    '''
    ursgal class
    '''
    def __init__(self, *args, **kwargs):
        # assert 'engine_path' in kwargs.keys(), \
        #     'Require engine_path as key/value input'
        # # try:
        #     self.exe   = kwargs['engine_path']
        # except:
        #     print(kwargs)
        #     print(self.engine)
        #     exit(1)
        self.params = {
            # 'executable_path' : self.exe,
            'file_info': {}
        }

        self.stats               = self._regenerate_default_stats()
        # This can be updated during __init__
        self.created_tmp_files   = []
        self.lookups             = {}
        self.user_defined_params = {}
        self.force               = False
        self.preflight_answer    = None
        self.execute_answer      = None
        self.postflight_answer   = None
        self.dependencies_ok     = True
        #self.loaded_json = False  # not used?

    def _regenerate_default_stats( self ):
        stats = {
            'time_points'    : {},
            'processed_files': {},
            'history'        : [],
            'startup'        : [],
            'status'         : 'idle'
        }
        return stats

    def abs_paths_for_specific_keys( self, params, param_keys = None):
        '''
        Absolute paths for specific keys from the params dict are determined

        Returns:
            dict: params with paths in abspath version
        '''
        if param_keys is None:
            if 'database' in params:
                param_keys = ['database']
            else:
                return params
        for san_key in param_keys:
            params[ san_key ] = os.path.abspath( params[ san_key ])
        return params

    def calc_md5( self, input_file):
        '''
        Calculated MD5 for input_file

        Args:

            input_file (str): Path to file

        Returns:

            str: MD5 of input file


        Thanks Raymond :)
        http://stackoverflow.com/questions/7829499/using-hashlib-to-compute-md5-digest-of-a-file-in-python-3
        '''
        self.print_info(
            'Calculating md5 for {0} ....'.format(
                os.path.basename(input_file),
                # tag = 'md5'
            ),
            caller='md5'
        )
        with open(input_file, mode='rb') as f:
            d = hashlib.md5()
            for buf in iter(partial(f.read, 1024), b''):
                d.update(buf)
        return d.hexdigest()

    # def check_if_all_default_params_are_in_params(self):
    #     for default_param_key, default_param_value in self.DEFAULT_PARAMS.items():
    #         if default_param_key not in self.params.keys():
    #             self.params[ default_param_key ] = default_param_value
    #     return

    def determine_common_top_level_folder( self, input_files=None ):
        '''
        The unode function determines for a list of input
        files a common top level folder they all belong to

        Keyword Arguments:
            input_files (list): list with input files

        Returns:
            str: The common top level folder
        '''
        paths = []
        for input_file in input_files:
            filepath, file_name = os.path.split(
                input_file
            )
            #print( filepath , input_file , os.path.abspath(filepath))
            splitted_path = filepath.split(
                os.sep
            )
            paths.append( ( len(splitted_path), splitted_path ) )
        paths.sort( reverse=True )
        common_top_level_folder = [
            os.path.abspath(os.sep)
        ]
        for pos in range( paths[0][0] ):
            folder_set = set()
            for number_of_path_elements, path in paths:
                if number_of_path_elements <= pos:
                    folder_set.add( None )
                else:
                    folder_set.add( path[ pos ] )
            if len( folder_set ) != 1:
                break
            else:
                common_top_level_folder.append( path[ pos ] )
        common_top_level_folder.append( '' )
        return os.path.join( *common_top_level_folder )


    def determine_common_name( self, input_files, mode=None):
        '''
        The unode function determines for a list of input
        files a basic common name

        Keyword Arguments:
            mode: head or tail for first or last part of the filename, respectively

        Arguments:
            input_files (list): list with input file names

        Returns:
            str: common file name
        '''
        if mode is None:
            mode = 'head'
        assert mode in ['head', 'tail'], '''
        Can only determine common part of a name from head or tail ...
        '''
        names = []
        for input_file in input_files:
            inf_basename = os.path.basename( input_file )
            if mode == 'head':
                inf_basename = inf_basename
            else:
                inf_basename = inf_basename[::-1]

            names.append( ( len(inf_basename), inf_basename ) )
        names.sort( reverse=True )
        longest_name_length = names[0][0]
        common_name = ''
        for pos in range( longest_name_length ):
            char_set = set()
            for name_length, name in names:
                if name_length <= pos:
                    char_set.add( None )
                else:
                    char_set.add( name[ pos ] )
            if len( char_set ) != 1:
                # print('Breaking ...')
                break
            else:
                common_name += char_set.pop()
            # print( pos , name[ pos ], len(common_name), len(char_set), common_name )
        return common_name

    def determine_common_head_tail_part_of_names( self, input_files=None):
        common_head = self.determine_common_name(
            input_files,
            mode='head'
        )
        common_tail = self.determine_common_name(
            input_files,
            mode='tail'
        )
        return common_head, common_tail[::-1]

    def update_output_json(self):
        '''
        Updates self.io['output']['params'] with self.io['input']['params']

        Although re-run might not be triggered, we need to update
        the output_json.
        '''
        number_of_params_changed = 0
        for i_param, i_value in self.io['input']['params'].items():
            if i_param not in self.io['output']['params'] or \
                    self.io['output']['params'][ i_param ] != i_value:
                number_of_params_changed += 1
                self.io['output']['params'][ i_param ] = i_value

        self.io['output']['finfo']['md5'] = self.io['output']['o_finfo']['md5']
        # self.io['output']['stats'] = self.io['output']['o_finfo']['stats']
        # print('{0} params changed'.format( number_of_params_changed ))
        # print('''

        #     Input

        # ''')
        # pprint.pprint( self.io['input'] )
        # print('''

        #     Output

        # ''')
        # pprint.pprint( self.io['output'] )
        # input('>>>>> Read you jsons! <<<<<<')
        self.dump_json_and_calc_md5( calc_md5 = False)
        return

    def dump_json_and_calc_md5(self, stats=None, params=None, calc_md5=True ):
        '''
        Dumps json with params and stats and calcs md5 for output

        Deletes all entries that are defined in
        params['del_from_params_before_json_dump']
        or keys that start with '_'
        '''
        self.print_info(
            'Preparing json dump',
            caller = 'dmpjson'
        )
        if stats is None:
            stats = self.io['output']['stats'].copy()
        if params is None:
            params = self.io['output']['params'].copy()
        json_file = os.path.join(
            self.io['output']['finfo']['dir'],
            self.io['output']['finfo']['json']
        )
        output_file = os.path.join(
            self.io['output']['finfo']['dir'],
            self.io['output']['finfo']['file']
        )
        # pprint.pprint( self.io['output'])
        # exit(1)
        if os.path.exists( output_file ) and calc_md5:
            self.io['output']['finfo']['md5'] = self.calc_md5(
                output_file
            )
            # print('''

            #     md5:{0}

            #     '''.format( self.io['output']['finfo']['md5']  ))
        j_content = [
            self.io['input']['finfo'].copy(),
            self.io['output']['finfo'].copy(),
            params,
            stats,
        ]
        keys_to_delete = self.params.get(
            'del_from_params_before_json_dump',
            []
        )
        for key in keys_to_delete:
            if key in j_content[2].keys():
                del j_content[2][ key ]

        for param_key in list(j_content[2].keys()):
            if param_key.startswith('_'):
                del j_content[2][ param_key ]

            elif isinstance(j_content[2][ param_key ], dict):
                for sub_param_key in list(j_content[2][ param_key ].keys()):
                    if sub_param_key.startswith('_'):
                        del j_content[2][ param_key ][sub_param_key]

        with open( json_file , 'w') as file_object:
            json.dump(
                j_content,
                file_object,
                sort_keys = True,
                indent = 2
            )
        self.print_info(
            'Json dumped. Path: {0}'.format(
                json_file
            ),
            caller='dmpjson'
        )
        return json_file


    def _execute(self):
        '''
        The _execute unode function

        Executes the unode executable via shell.

        Note: internal function
            Unodes that do not require execution via shell
            redefine the _execute() function in their engine
            class.

        Returns:
            None
        '''
        self.print_info('Executing command list ...', caller='eXecution')
        assert 'command_list' in self.params.keys(), '''
  No command_list was found in self.params. Convention is to define
  the command list during preflight in the uNode Engine class code or,
  alternatively, redefine uNode._execute() in the engine class altogether
        '''
        self.time_point(tag = 'execution')
        execute_answer = []
        proc = None

        if len(self.params['command_list']) != 0:
            mswindows = ( sys.platform == "win32" )
            if mswindows is True:
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                startupinfo.wShowWindow = subprocess.SW_HIDE
                proc = subprocess.Popen(
                    self.params['command_list'],
                    stdout = subprocess.PIPE,
                    shell = False,
                    startupinfo=startupinfo,
                )
            else:
                proc = subprocess.Popen(
                    self.params['command_list'],
                    stdout = subprocess.PIPE,
                    # shell = True
                )
        else:
            print('Command list is empty, nothing to do here...')
            print('_execute failed ....', self.params['command_list'])
            execute_answer.append( 'Command list is empty' )
            self.execute_return_code = 500

        if proc is not None:
            for line in proc.stdout:
                line_decoded = line.strip().decode('utf')
                print( line_decoded )
                execute_answer.append( line_decoded )

            # catching the executable's exit code to detect crashes:
            proc.communicate()[0]
            self.execute_return_code = proc.returncode
            assert self.execute_return_code in [0, None], '''
  \n{0} crashed!

  The executable
    {1}
  terminated with Error code {2} .
  Inspect the printouts above for possible causes and verify that all input files are valid.
            '''.format( self.engine, os.path.relpath(self.exe), self.execute_return_code)

        self.print_execution_time(tag='execution')
        return execute_answer


    def generate_fasta_database_lookup( self, database ):
        if database not in self.lookups['fasta_dbs'].keys():
            self.lookups['fasta_dbs'][ database ] = {}
            lookup_hook = self.lookups['fasta_dbs'][ database ]
            self.print_info(
                'Fasta_db {0}'.format( database ),
                caller ='Caching'
            )
            current_id = None
            for line in open( database ):
                if line[:1] == '>':
                    if line.strip() in lookup_hook.keys():
                        print('[ WARNING ] Fasta database contains duplicates !')
                    current_id = line[1:].strip()
                    lookup_hook[ current_id ] = ''
                else:
                    lookup_hook[ current_id ] += line.strip()
            self.print_info(
                'Done!',
                caller ='Caching'
            )

    def flatten_list(self, multi_list=[]):
        '''
        The unode get_last_engine function

        Reduces a multidimensional list of lists to a flat list including all elements
        '''
        if multi_list == []:
            return multi_list
        if isinstance(multi_list[0], list):
            return self.flatten_list(multi_list[0]) + self.flatten_list(multi_list[1:])
        return multi_list[:1] + self.flatten_list(multi_list[1:])

    def get_last_engine(self, history=None, engine_types=None, multiple_engines=False):
        '''
        The unode get_last_engine function

        Note: returns None if the specified engine type was not used yet.

        Keyword Arguments:
            history (list): A list of path unodes, timestamps and parameters
                that were used. This function can be used on the history loaded
                from a file .json, to find out which search engine was used on
                that file.
                If not specified, this information is taken from the unode class
                itself, and not a specific file.
            engine_types (list): the engine type(s) for which the last used engine
                should be identified
            multiple_engines (bool): if muliple engines have been used, this can be
                set to True. Then reports a list of used engines.

        Examples:
            >>> fpaths = self.generate_basic_file_info( "14N_xtandem.csv" )
            >>> file_info, __    = self.load_json( fpaths=fpaths, mode='input')
            >>> last_engine      = self.get_last_engine(
                    history      = file_info["history"],
                    engine_types = ["protein_database_search_engine"]
                )
            >>> print( last_engine )
            "xtandem_sledgehammer"

        Returns:
            str: The name of the last engine that was used.
        '''
        last_engine = None
        if engine_types is None:
            engine_types = [
                'converter',
                'cross_link_search_engine',
                'de_novo_search_engine',
                'fetcher',
                'misc_engine',
                'meta_engine',
                'protein_database_search_engine',
                'spectral_library_search_engine',
                'validation_engine',
                'visualizer',
            ]
        if not history:
            history = self.stats["history"]
        else:
            for history_event in history[::-1]:
                if history_event['engine'] == 'merge_csvs_1_0_0':
                    merged_engines = set()
                    for element in self.flatten_list(history_event['history_addon']['search_engines_of_merged_files']):
                        merged_engines.add(element)
                    if len( merged_engines ) == 1:
                        last_engine = list(merged_engines)[0]
                    elif multiple_engines == True:
                        last_engine = list(merged_engines)

                    else:
                        assert last_engine != None, '''
                        last_engine cannot be determined, since multiple engines have been used.
                        If you want to return all used engines, set multiple_engines=True
                        {0}
                        '''.format( ", ".join(merged_engines) )
                    break

                meta = history_event.get("META_INFO", {})
                meta_engine_type_info = meta.get("engine_type", {})
                for engine_type in engine_types:
                    if meta_engine_type_info.get( engine_type, False ):
                        last_engine = history_event["engine"]
                        break

        return last_engine

    def get_last_engine_type(self, history=None):
        '''
        The unode get_last_engine_type function

        Keyword Arguments:
            history (list): A list of path unodes, timestamps and parameters
                that were used. This function can be used on the history loaded
                from a file .json, to find out which search engine was used on
                that file.
                If not specified, this information is taken from the unode class
                itself, and not a specific file.

        Examples:
            >>> fpaths = self.generate_basic_file_info( "14N_xtandem.csv" )
            >>> file_info, __    = self.load_json( fpaths=fpaths, mode='input')
            >>> last_engine_type = self.get_last_engine_type(
                    history      = file_info["history"],
                )
            >>> print( last_engine_type )
            "protein_database_search_engine"

        Returns:
            str: The type of the last engine that was used.
            Returns None if the engine_type cannot be specified or if no engine
            was previously executed on this file.
        '''

        last_engine_type = None
        if not history:
            history = self.stats["history"]
        history_event = history[-1]
        meta = history_event.get("META_INFO", {})
        meta_engine_type_info = meta.get("engine_type", {})

        for engine_type in meta_engine_type_info.keys():
            if meta_engine_type_info.get( engine_type, False ):
                last_engine_type = engine_type
                break

        return last_engine_type

    def get_last_search_engine(self, history=None, multiple_engines=False ):
        '''
        The unode get_last_search_engine function

        Note: returns None if no search engine was not used yet.

        Keyword Arguments:
            history (list): A list of path unodes, timestamps and parameters
                that were used. This function can be used on the history loaded
                from a file .json, to find out which search engine was used on
                that file.
                If not specified, this information is taken from the unode class
                itself, and not a specific file.
            multiple_engines (bool): if muliple engines have been used, this can be
                set to True. Then reports a list of used engines.

        Examples:
            >>> fpaths = self.generate_basic_file_info( "14N_xtandem.csv" )
            >>> file_info, __ = self.load_json( fpaths=fpaths, mode='input')
            >>> last_engine   = self.get_last_search_engine(
                    history   = file_info["history"]
                )
            >>> print( last_engine )
            "xtandem_sledgehammer"

        Returns:
            str: The name of the last search engine that was
            used. Returns None if no search engine was used yet.
        '''
        last_search_engine = None
        last_search_engine = self.get_last_engine(
            history=history,
            engine_types=[
                'cross_link_search_engine',
                'de_novo_search_engine',
                'protein_database_search_engine',
                'spectral_library_search_engine',
            ],
            multiple_engines=multiple_engines,
        )
        return last_search_engine

    def _group_psms(self, input_file, validation_score_field=None, bigger_scores_better=None):
        '''
        Reads an input csv and returns a defaultdict with the spectrum title
        mapping to a sorted list of tuples containing each
        a) score (from validation_score_field) and
        b) the whole line dict

        Keyword Arguments:
            validation_score_field (str): fieldname of the column that should be used as validation score
                for sorting of PSMs. If None, get_last_search_engine is used to get the validation_score_field
                defined for the last used search engine.
            bigger_scores_better (bool): defines if in the validation score are increasing (True) or decreasing
                (False) with their quality. If None, get_last_search_engine is used to get  bigger_scores_better
                defined for the last used search engine.
        '''
        print('[ GROUPING ] Parsing {0}'.format( input_file ))

        if bigger_scores_better is None or validation_score_field is None:
            last_engine = self.get_last_search_engine(
                history=self.stats['history']
            )
            assert last_engine, 'Can\'t convert results from no specified search engine.'
            assert 'multiple engines:' not in last_engine, 'Can\'t convert merged results from multiple different engines.'
            if bigger_scores_better is not None or validation_score_field is not None:
                print('''
[ WARNING ] Only "validation_score_field" or "bigger_scores_better" was defined
[ WARNING ] in UController.params. Determining both (!)
[ WARNING ] automatcally based on the last engine: {},
                '''.format(last_engine))
            if last_engine is None or type(last_engine) == list:
                print('''
                    Could not determine last_search_engine from input file.
                    Got {0}
                    Please specify parameters validation_score_field and bigger_scores_better
                '''.format(last_engine))
                sys.exit(1)
            else:
                bigger_scores_better = self.UNODE_UPARAMS['bigger_scores_better']['uvalue_style_translation'][last_engine]
                validation_score_field = self.UNODE_UPARAMS['validation_score_field']['uvalue_style_translation'][last_engine]

        tmp = ddict(list)
        grouped_psms = ddict(list)

        opened_file = open( input_file, 'r')
        csv_dict_reader_object = csv.DictReader(
            row for row in opened_file if not row.startswith('#')
        )
        n = 0

        for n, line_dict in enumerate(csv_dict_reader_object):
            assert validation_score_field in line_dict.keys(), \
                '''defined validation_score_field {0} is not found,
                please check/add it to uparams.py['validation_score_field']'''.format(validation_score_field)

            tmp[ line_dict[ 'Spectrum Title' ] ].append(
                (
                    float(
                        line_dict[ validation_score_field ]
                    ),
                    line_dict
                )
            )

        for spectrum_title in tmp.keys():
            already_seen = set()

            for score, line_dict in sorted( tmp[ spectrum_title ],
                    key     = operator.itemgetter(0),
                    reverse = bigger_scores_better
                ):
                _identifier = '_'.join( sorted( line_dict.values() ))
                if _identifier in already_seen:
                    continue
                already_seen.add( _identifier )
                grouped_psms[ spectrum_title ].append( ( score, line_dict ) )

        print(
            "[ GROUPING ] Grouped {0} PSMs into {1} unique spectrum titles".format(
                n,
                len( grouped_psms.keys())
            )
        )
        return grouped_psms

    def import_engine_as_python_function( self, function_name = None ):
        '''
        The unode import_engine_as_python_function function

        Imports the main function from a unodes "executable". For unodes
        that are written completely in python and can be executed by
        importing them instead of using the command line.

        Examples:
            >>> us = ursgal.UController()
            >>> cFDR_unode = us.unodes["combine_FDR_0_1"]["class"]
            >>> cFDR_main  = cFDR_unode.import_engine_as_python_function()
            >>> cFDR_main(
                input_file_list = ["1.csv", "2.csv"],
                directory       = "/tmp/",
            )

        Returns:
            function: The function called "main" that
            is specified in the engines python script.

        Note:
            Assertion exception if the executable is not a python
            script, or has no main function.
        '''
        assert os.path.exists( self.exe ), '''
        Engine needs to be python source file in order to import its main
        function!
        '''
        module_dir_name = os.path.dirname( self.exe )
        sys.path.insert( 1, module_dir_name )
        imported_module = importlib.import_module( self.engine )
        if function_name is None:
            function_name = 'main'
        assert hasattr(imported_module, function_name), '''
        Can not import main() function from engine
        {0}
        '''.format( self.exe )
        main_function = getattr( imported_module, function_name)
        return main_function

    def load_json( self, finfo=None, json_path=None):
        j_content = None
        if json_path is None:
            json_path = os.path.join( finfo['dir'], finfo['json'] )
        if os.path.exists( json_path ):
            j_content = json.load( open( json_path, 'r' ) )

        self.compare_json_and_local_ursgal_version(
            history = j_content[3]['history'],
            json_path = json_path
        )
        return j_content

    def compare_json_and_local_ursgal_version(self, history, json_path):
        ''' Print a warning if the history is a from a different version number '''
        json_version = None
        if len(history) > 0 and 'ursgal_version' in history[-1]:
            json_version = history[-1]['ursgal_version']
        if json_version != None and json_version != ursgal.__version__:
            self.print_info(
                ("JSON {0} stems from another Ursgal version!"
                    ).format(os.path.relpath(json_path)), caller="/\\")
            self.print_info(
                ("JSON version: {0} - Local version: {1}"
                    ).format(json_version, ursgal.__version__),
                caller="WARNING!")
            self.print_info(
                "Proceed at your own risk, or use (the) force to re-run the whole analysis!",
                caller="/____\\")
        return

    def map_mods(self):
        '''Maps modifications defined in params["modification"] using unimod.

       Examples:

            >>> [
            ...    "M,opt,any,Oxidation",        # Met oxidation
            ...    "C,fix,any,Carbamidomethyl",  # Carbamidomethylation
            ...    "*,opt,Prot-N-term,Acetyl"    # N-Acteylation
            ... ]

        '''
        self.params[ 'mods' ] = {
            'fix' : [],
            'opt' : []
        }
        for ursgal_index, mod in enumerate(
                sorted( self.params[ 'modifications' ] )):
            mod_params  = mod.split( ',' )
            if len(mod_params) >=6 or len(mod_params) <=3:
                print('''
    [ WARNING ] For modifications, please use the ursgal_style:
    [ WARNING ] 'amino_acid,opt/fix,position,Unimod PSI-MS Name'
    [ WARNING ] or
    [ WARNING ] 'amino_acid,opt/fix,position,name,chemical_composition'
    [ WARNING ] Continue without modification {0} '''.format( mod )
                )
                print(mod_params)
                continue
            aa          = mod_params[ 0 ].strip()
            mod_option  = mod_params[ 1 ].strip()
            pos         = mod_params[ 2 ].strip()
            unimod = False
            unimod_id = None

            if len(mod_params) == 4:
                try:
                    unimod_id = int(mod_params[ 3 ].strip())
                    unimod_name   = ursgal.GlobalUnimodMapper.id2name( unimod_id )
                    mass = ursgal.GlobalUnimodMapper.id2mass( unimod_id )
                    composition = ursgal.GlobalUnimodMapper.id2composition( unimod_id )
                    if unimod_name is None:
                        print('''
    [ WARNING ] '{1}' is not a Unimod modification
    [ WARNING ] please change it to a valid Unimod Accession # or PSI-MS Unimod Name
    [ WARNING ] or add the chemical composition hill notation (including 1)
    [ WARNING ] e.g.: H-1N1O2
    [ WARNING ] ursgal_style: 'amino_acid,opt/fix,position,name,chemical_composition'
    [ WARNING ] Continue without modification {0} '''.format(
                            mod,
                            unimod_id
                        ))
                        continue
                    unimod = True
                    name = unimod_name
                except:
                    unimod_name = mod_params[ 3 ].strip()
                    unimod_id   = ursgal.GlobalUnimodMapper.name2id( unimod_name )
                    mass = ursgal.GlobalUnimodMapper.name2mass( unimod_name )
                    composition = ursgal.GlobalUnimodMapper.name2composition( unimod_name )
                    if unimod_id is None:
                        print('''
    [ WARNING ] '{1}' is not a Unimod modification
    [ WARNING ] please change it to a valid PSI-MS Unimod Name or Unimod Accession #
    [ WARNING ] or add the chemical composition hill notation (including 1)
    [ WARNING ] e.g.: H-1N1O2
    [ WARNING ] ursgal_style: 'amino_acid,opt/fix,position,name,chemical_composition'
    [ WARNING ] Continue without modification {0} '''.format(
                            mod,
                            unimod_name
                        ))
                        continue
                    unimod = True
                    name = unimod_name

            elif len(mod_params) == 5:
                name = mod_params[ 3 ].strip()
                chemical_formula = mod_params[ 4 ].strip()
                chemical_composition = ursgal.ChemicalComposition()
                chemical_composition.add_chemical_formula( chemical_formula )
                composition = chemical_composition
                composition_unimod_style = chemical_composition.hill_notation_unimod()
                unimod_name_list = ursgal.GlobalUnimodMapper.composition2name_list( composition_unimod_style )
                unimod_id_list = ursgal.GlobalUnimodMapper.composition2id_list( composition_unimod_style )
                mass = ursgal.GlobalUnimodMapper.composition2mass( composition_unimod_style )
                for i, unimod_name in enumerate(unimod_name_list):
                    if unimod_name == name:
                        unimod_id = unimod_id_list[ i ]
                        unimod = True
                        break
                if unimod == False and unimod_name_list != []:
                    print(    '''
    [ WARNING ] '{0}' is not a Unimod modification
    [ WARNING ] but the chemical composition you specified is included in Unimod.
    [ WARNING ] Please use one of the Unimod names:
    [ WARNING ] {1}
    [ WARNING ] Continue without modification {2} '''.format(
                            name,
                            unimod_name_list,
                            mod
                    ))
                    continue
                if unimod == False and unimod_name_list == []:
                    print (    '''
    [ WARNING ] '{0}' is not a Unimod modification
    [ WARNING ] trying to continue with the chemical composition you specified
    [ WARNING ] This is not working with OMSSA so far'''.format(
                            mod,
                    ))
                    mass = chemical_composition._mass()
                    # write new userdefined modifications Xml in unimod style

            mod_dict = {
                '_id'   : ursgal_index,
                'aa'    : aa,
                'mass'  : mass,
                'pos'   : pos,
                'name'  : name,
                'composition' : composition,
                'org'   : mod,
                'id'    : unimod_id,
                'unimod': unimod,
            }
            if mod_dict['unimod'] == False:
                ursgal.GlobalUnimodMapper.writeXML( mod_dict )

            self.params[ 'mods' ][ mod_option ].append( mod_dict )
        return self.params[ 'mods' ]

    def peptide_regex(self, database, protein_id, peptide):
        '''
        Note:
            This function is not longer used at the moment.

        The unode peptide_regex function

        Args:
            database (str): Name of the used fasta database
            protein_id (str): protein ID of the processed protein
            peptide (str): peptide which should be mapped on the protein ID's
                sequence

        This function takes a peptide sequence and maps it to its according
        proteins sequence, returning the start and stop position in the sequence
        as well as the amino acid before and after the peptide sequence in the
        full protein sequence. If the peptide sequence contains known amino acid
        substitutions like U (Selenocystein) or J (Leucin or Isoleucin) this
        amino acid is replaced by a regex wildcard '.' in order to be matchable
        on the fasta database (this is defined in kb.unify_csv_1_0_0.py).
        This is especially needed if the original sequence contains a 'X' and
        the search engine guesses/determines the amino acid at this position.

        If the protein ID is ambigous, the peptide is matched against all
        protein candidates and the positions, pre- and post aminoacids in the
        matching sequence as well as the full protein ID as named in the fasta
        database is returned. This is especially needed for MS Amanda results
        where protein IDs are returned truncated and become ambigous for some
        databases.

        If the peptide occurs several times in the protein, all occurences are
        returned.

        The function uses a buffer to perform the regex only once for
        (peptide, protein, database) tuples. All fasta sequences are also
        buffered in `self.lookups['fasta_dbs']` with the name of the database as
        key and then all protein IDs and sequnces as key, value pairs.

        Pre and post amino acids are required for e.g. percolator input files.

        Note:

            The regex and peptide to protein ID mapping may take a while, if a
            large file has to be processed.

        Returns:
            list: list of tuples
            [( peptide_start, peptide_stop, aa_before_peptide, aa_after_peptide, protein_id )]


        '''
        if 'fasta_dbs' not in self.lookups.keys():
            self.lookups['fasta_dbs'] = {}
        if database not in self.lookups['fasta_dbs'].keys():
            self.generate_fasta_database_lookup( database )
        # print(self.lookups)
        # unify_csv_converter_version = self.params['unify_csv_converter_version']
        peptide_for_regex = peptide
        # corrected_peptide = peptide
        for aa_exception, aa_exception_info in self.meta_unodes[ 'unify_csv_1_0_0' ].DEFAULT_PARAMS['aa_exception_dict'].items():
            #regex will match any character in place of unknown aminoc acid...
            peptide_for_regex = peptide_for_regex.replace(
                aa_exception,
                '.'
            )
        ids_matched = []
        if protein_id not in self.lookups['fasta_dbs'][ database ].keys():
            #ms amanda short name problem
            for long_protein_id in self.lookups['fasta_dbs'][ database ].keys():
                if long_protein_id.startswith( protein_id ):
                    sequence_to_check = self.lookups['fasta_dbs'][ database ][long_protein_id]
                    match = re.search( peptide_for_regex, sequence_to_check )
                    if match is not None:
                        ids_matched.append( long_protein_id )
        else:
            ids_matched = [ protein_id ]
        return_buffer_keys = []
        for protein_id in ids_matched:
            full_sequence = self.lookups['fasta_dbs'][ database ][ protein_id ]
            buffer_key = ( database, peptide_for_regex, protein_id )
            if 'peptide_pos_buffer' not in self.lookups.keys():
                self.lookups['peptide_pos_buffer'] = {}
            if buffer_key not in self.lookups['peptide_pos_buffer'].keys():
                match_found = False
                for match in re.finditer( peptide_for_regex, full_sequence ):
                    match_found = True
                    start = match.start()
                    stop = start + len(peptide)

                    if start == 0:
                        pre_aa = '-'
                    else:
                        pre_aa = full_sequence[ start - 1 ]

                    if stop == len(full_sequence):
                        post_aa = '-'
                    else:
                        post_aa = full_sequence[ stop ]
                    pos_aa_protein = (
                        start + 1,
                        stop,
                        pre_aa,
                        post_aa,
                        protein_id
                        )
                    if buffer_key in self.lookups['peptide_pos_buffer']:
                        self.lookups['peptide_pos_buffer'][buffer_key].append(pos_aa_protein)
                    else:
                        self.lookups['peptide_pos_buffer'][buffer_key] = [pos_aa_protein]
                if match_found == False:
                    self.lookups['peptide_pos_buffer'][buffer_key] = [(
                        None,
                        None,
                        None,
                        None,
                        protein_id
                        )]
            # print(self.lookups['peptide_pos_buffer'].keys())
            if self.lookups['peptide_pos_buffer'][buffer_key][0] is None:
                self.print_info(
                    'Peptide {0} cant be found in database {1}, peptide for regex: {2}'.format(
                        peptide,
                        self.params['database'],
                        peptide_for_regex,
                    ),
                    caller='Warning'
                )

            return_buffer_keys.append( self.lookups['peptide_pos_buffer'][buffer_key] )
        return return_buffer_keys

    def _preflight(self):

        if 'citation' in self.META_INFO:
            self.print_info( caller='Citation', msg='')
            self.print_info( caller='Please', msg='')
            self.print_info(
                self.META_INFO["citation"],
                caller='cite:'
            )
            self.print_info( caller='-----', msg='')
            self.print_info( caller='Citation', msg='')

        self.print_info(
            'Executing preflight sequence ...',
            caller='PREFLGHT'
        )
        self.time_point(tag = 'preflight')
        preflight_answer = self.preflight()
        self.print_execution_time(tag='preflight')
        return preflight_answer

    def preflight(self):
        '''This can be/is overwritten by the engine uNode class'''
        return

    def _postflight(self):
        self.print_info(
            'Executing postflight sequence ...',
            caller = 'POSTFLGHT'
        )
        self.time_point(tag = 'postflight')
        postflight_answer = self.postflight()
        self.print_execution_time(tag='postflight')
        return postflight_answer

    def postflight(self):
        '''This can be/is overwritten by the engine uNode class'''
        return

    def print_execution_time( self, tag=None ):

        lap = self.time_point(tag = tag, diff=True, stop=True)
        # print('>>>', tag )
        # print('>>>', self.stats['time_points'])
        # print('>>>>', lap )
        # print('>>>>', self.time_point(tag = 'execution'))
        if lap > 3600:
            unit = 'hours'
            lap /= 3600
        elif lap > 60:
            unit = 'minutes'
            lap /= 60
        else:
            unit = 'seconds'
        msg = 'Execution time {0:.3f} {1}'.format(
            lap,
            unit
        )
        self.print_info( msg, caller=tag )

    @classmethod
    def print_info( cls, msg, caller=None ):
        if caller is None:
            if hasattr(cls, 'engine'):
                caller = cls.engine
            else:
                caller = ''
        if len(caller) > 7:
            caller = caller[:8]

        print(
            '[ {0: ^08s} ] {1}'.format(
                caller,
                msg
            )
        )

    def print_header( self, header, tag=None, newline=True):
        if tag is not None:
            time_stamp = '({0})'.format(
                self.time_point(
                    tag         = tag,
                    diff        = False,
                    format_time = True,
                    stop        = False
                )
            )
        else:
            time_stamp = ''
        print('''
        -\-     {0} {1}     -/-'''.format(
            header,
            time_stamp
        ))
        if newline:
            print()

    def initialize_run(self ):
        self.created_tmp_files = []
        self.io = {
            'input'  : {
                'finfo'  : None,
                'params' : None,
                'stats'  : None
            },
            'output' : {
                'finfo'  : None,
                'params' : None,
                'stats'  : None
            },
        }
        self.preflight_return_code = 0
        self.execute_return_code = 0
        self.postflight_return_code = 0
        self.stats = {}

    def generate_empty_report( self ):
        empty_report = {
            'preflight' : None,
            'execution' : None,
            'postflight': None,
            'stats' : None,
            'addon' : None,
        }
        return empty_report

    def update_params_with_io_data(self):
        '''
        Generates a flat structure in params combining
        io.['input']['finfo'] &
        io.['output']['finfo']
        '''
        i_finfo = self.io['input']['finfo']
        self.params['input_dir_path'] = i_finfo['dir']
        self.params['input_file'] = i_finfo['file']
        self.params['file_root'] = i_finfo['file_root']
        self.params['input_is_compressed'] = i_finfo['is_compressed']

        o_finfo = self.io['output']['finfo']
        self.params['output_dir_path'] = o_finfo['dir']
        self.params['output_file'] = o_finfo['file']

    def fix_md5_and_file_in_json(self, ifile=None, json_path=None):
        '''
        Fixes supplementary output json
        '''
        self.print_info(
            'Preparing json fix',
            caller = 'fix_md5'
        )
        if ifile is None:
            ifile = json_path.replace('.u.json', '')
            if os.path.exists(ifile) is False:
                raise IOError('file corresponding to {0} does not exists'.format(json_path))
        elif json_path is None:
            json_path = ifile+'.u.json'
            if os.path.exists(json_path) is False:
                raise IOError('json ({0}) corresponding to {1} does not exists'.format(json_path, ifile))
        i_finfo, o_finfo, params, stats = self.load_json(json_path=json_path)

        file_root, file_extention = os.path.splitext(os.path.basename(ifile))
        new_o_finfo = {
            'dir': os.path.dirname(ifile),
            'file': os.path.basename(ifile),
            'file_root': file_root,
            'file_extention': file_extention,
            'full': ifile,
            'is_compressed': False,
             'json': os.path.basename(json_path),
             'json_exists': True,
             'md5' : self.calc_md5(ifile)
        }

        j_content = [
            i_finfo,
            new_o_finfo,
            params,
            stats,
        ]
        with open(json_path , 'w') as file_object:
            json.dump(
                j_content,
                file_object,
                sort_keys = True,
                indent = 2
            )
        self.print_info(
            'New json dumped {0}'.format(os.path.basename(json_path)),
            caller = 'fix_md5'
        )


    def run(self, json_path=None ):
        '''
        The general run function.

        Runs engine/uNode child with given params on defined input_file.
        This function is automatically called by all ucontroller functions
        that take an input file and produce a single output file (i.e.
        ucontroller.search() and ucontroller.validate() )


        Keyword Arguments:
            json_path (str): path to input file json, dumped by a controller
            # input_file (str): path to the input file
            # fpaths (dict): dictionary containing file path information.
            # If None, this is generated using unode.generate_basic_file_info( file )
            # force (bool): (re)do the analysis if ouput files already exists

        Returns:
            dict: Report of the run.

        Note:

            Internal function. This function executes the preflight, postflight
            and _execute functions, if defined in the engine python script.

        '''
        self.initialize_run()
        i_finfo, o_finfo, params, stats = self.load_json( json_path=json_path )
        self.params = params
        self.stats = stats
        self.io['input']['finfo'] = i_finfo
        self.io['output']['finfo'] = o_finfo
        self.update_params_with_io_data()
        report = self.generate_empty_report()

        tag = '{0}@{1}'.format(
            self.engine,
            self.io['input']['finfo']['full'],
        )
        self.time_point( tag = tag )
        self.print_header(
            '{0} run initialized with {1}'.format(
                self.engine,
                self.io['input']['finfo']['file']
            ),
            tag=tag
        )
        if self.io['output']['finfo']['is_compressed']:

            self.params['output_file'] = self.params['output_file'].replace('.gz', '')

            self.print_info(
                'Will compress output {output_file} on the fly ... renamed temporarily params["output_file"]'
            .format(
                **self.params
                ),
                caller = 'run'
            )

        # DEFAULT PARAMS ARE INCLUDED HERE :)
        # self.params = self.DEFAULT_PARAMS.copy()

        # self.check_if_all_default_params_are_in_params()

        # self.time_point(tag = 'run')
        self.stats['history'] = self.update_history_status(
            history = self.stats['history']
        )
        self.print_info(
            'Preparing engine',
            caller = 'run'
        )

        # We use translated in the first json dump
        # so we can check which parameters are used during
        # run time or which one lead to a crash

        # self.original_params = copy.deepcopy( self.params )


        # We use original params for second and final dump
        #
        # if 'command_list' in self.original_params:
        #     del self.original_params['command_list']

        if 'translations' in self.params:
            del self.params['translations']

        untranslated_params, translated_params = \
            self.collect_and_translate_params( self.params )
        self.params.update( untranslated_params )
        self.params['translations'] = translated_params

        is_search_engine = False
        for engine_type in ukb.ENGINE_TYPES.keys():
            if 'search_engine' in engine_type:
                is_search_engine = self.META_INFO['engine_type'].get(
                    engine_type,
                    False
                )
            if is_search_engine is True:
                break
        # is_search_engine = self.META_INFO['engine_type'].get(
        #     'search_engine',
        #     False
        # )
        # is_denovo_engine = self.META_INFO['engine_type'].get(
        #     'denovo_engine',
        #     False
        # )
        # is_crosslink_engine = self.META_INFO['engine_type'].get(
        #     'cross_link_engine',
        #     False
        # )
        is_quantification_engine = self.META_INFO['engine_type'].get(
            'quantification_engine',
            False
        )
        map_mods_node_exceptions = [
            'unify_csv'
        ]
        if is_search_engine or is_quantification_engine:
            self.map_mods()
        for engine_short_name in map_mods_node_exceptions:
            if engine_short_name in self.engine:
                self.map_mods()
                break
        self.stats['history'] = self.update_history_status(
            status='launching',
            history = self.stats['history']
        )

        requires_grouped_psms = self.META_INFO.get(
            'group_psms',
            False
        )
        if requires_grouped_psms:
            self.print_info(
                'Grouping PSMs',
                caller ='run'
            )
            self.time_point(tag='group_psms')
            self.params['grouped_psms'] = self._group_psms(
                os.path.join(
                    self.params['input_dir_path'],
                    self.params['input_file']
                ),
                validation_score_field = self.params['translations']['validation_score_field'],
                bigger_scores_better = self.params['translations']['bigger_scores_better']
            )
            self.print_execution_time(tag = 'group_psms')

        self.print_info(
            'Starting engine',
            caller = 'run'
        )
        report['preflight'] = self._preflight()
        report['execution'] = self._execute()
        report['postflight'] = self._postflight()

        if self.execute_return_code == 0:
            history_status = 'complete'
        else:
            history_status = self.execute_return_code

        self.stats['history'] = self.update_history_status(
            status = history_status,
            history = self.stats['history']
        )
        # We Use original params for second and final dump
        # but keep novel params that are passed from the unode
        # jsons will get bigger if junk is collected
        # NOTE: all keys starting with _ will be deleted before dump
        if self.io['output']['finfo']['is_compressed']:
            gz_output_file = os.path.join(
                self.io['output']['finfo']['dir'],
                self.io['output']['finfo']['file']
            )
            raw_output = os.path.join(
                self.params['output_dir_path'],
                self.params['output_file'],
                )
            self.print_info('Compressing output {0} > {1}'.format(
                os.path.basename(raw_output),
                os.path.basename(gz_output_file),
            ))
            with open( raw_output, 'rb') as f_in:
                with gzip.open( gz_output_file, 'wb') as f_out:
                    f_out.writelines(f_in)
            self.created_tmp_files.append( raw_output )

        # self.params.update( self.original_params )

        if self.params['remove_temporary_files']:
            for tmp_file in self.created_tmp_files:
                if os.path.exists(tmp_file):
                    os.remove( tmp_file )

        self.print_execution_time(tag = tag)

        report['output_file'] = os.path.join(
            self.io['output']['finfo']['dir'],
            self.io['output']['finfo']['file']
        )

        # report['output_md5'] = self.calc_md5(
        #     report['output_file']
        # )
        report['stats'] = self.stats
        report['params'] = self.params
        return report

    def set_params( self, params):
        self.params.update( params )
        return


    def collect_and_translate_params(self, params):
        '''
        Translates ursgal parameters into uNode specific syntax.

        1) Each unode.USED_SEARCH_PARAMS contains params that have
            to be passed to the uNode.
        2) params values are not translated is they [] or {}
        3) params values are translated using::

            uNode.USEARCH_PARAM_VALUE_TRANSLATIONS
            > translating only values, regardless of key
            uNode.USEARCH_PARAM_KEY_VALUE_TRANSLATOR
            > translating only key:value pairs to key:newValue

        Those lookups are found in kb/{engine}.py

        TAG:
            - v0.4
        '''
        # print( self.DEFAULT_PARAMS )
        translated_params = {}
        GROUPED_TRANSLATIONS = {}
        untranslated_params = {}
        for ukey, mDict in self.UNODE_UPARAMS.items():

            pvalue = params.get( ukey, mDict['default_value'] )
            ukey_t = mDict['ukey_translated']

            if pvalue == mDict['default_value']:
                translated_value = mDict['default_value_translated']
            else:
                if isinstance( pvalue, list):
                    translated_value = pvalue
                elif isinstance( pvalue, dict):
                    translated_value = pvalue
                else:
                    translated_value = mDict['uvalue_style_translation'].get(
                        pvalue,
                        pvalue
                    )

            untranslated_params[ ukey ] = pvalue

            translated_params[ mDict['ukey'] ] = translated_value
            translated_params[ '{ukey}_key'.format(**mDict) ] = ukey_t

            if ukey_t not in GROUPED_TRANSLATIONS.keys():
                GROUPED_TRANSLATIONS[ ukey_t ] = {}
            GROUPED_TRANSLATIONS[ ukey_t ].update(
                {
                    mDict['ukey'] : translated_value
                }
            )
        translated_params['_grouped_by_translated_key'] = \
            GROUPED_TRANSLATIONS

        return untranslated_params, translated_params


        # exit("Translation is not done yet ....")
        # translated_params = {}
        # for mapped_unode_param in self.USED_USEARCH_PARAMS:
        #     assert mapped_unode_param in list(params.keys()), '''
        #     Missing required parameter {0}.
        #     All engine parameters (in kb.{1}.USED_USEARCH_PARAMS)
        #     have to be passed in params.
        #     '''.format(mapped_unode_param, self.engine, )

        #     ursgal_value = params[ mapped_unode_param ]
        #     # print(ursgal_value, type( ursgal_value ))
        #     if isinstance( ursgal_value, list) is True:
        #         translated_value = ursgal_value
        #     elif isinstance( ursgal_value, dict) is True:
        #         translated_value = ursgal_value
        #         # print( '''>>?<''' )
        #     else:
        #         # -- USEARCH_PARAM_KEY_VALUE_TRANSLATOR --
        #         LOOKUP = self.USEARCH_PARAM_KEY_VALUE_TRANSLATOR
        #         if mapped_unode_param in LOOKUP.keys():

        #             assert ursgal_value in LOOKUP[ mapped_unode_param ].keys(), '''
        # {0} is defined in kb.{1}.USEARCH_PARAM_KEY_VALUE_TRANSLATOR,
        # yet the ursgal_value {2}
        # has no specific translation defined.'''.format(
        #                 mapped_unode_param,
        #                 self.engine,
        #                 ursgal_value
        #             )
        #             translated_value = \
        #                 LOOKUP[ mapped_unode_param ][ ursgal_value ]
        #         # -- USEARCH_PARAM_VALUE_TRANSLATIONS --
        #         else:
        #             LOOKUP = self.USEARCH_PARAM_VALUE_TRANSLATIONS
        #             if ursgal_value in LOOKUP.keys():
        #                 lookup_list = list( LOOKUP.keys() )
        #                 index       = lookup_list.index( ursgal_value )
        #                 lookup_key  = lookup_list[ index ]
        #                 # print(ursgal_value, lookup_key)

        #                 if lookup_key == ursgal_value and type(lookup_key) == type(ursgal_value):
        #                     # print(ursgal_value, lookup_key)

        #                     translated_value = LOOKUP.get(ursgal_value, None)
        #                     # print(translated_value)
        #                     if translated_value is None:
        #                         translated_value = ursgal_value
        #                 else:
        #                     translated_value = ursgal_value
        #             else:
        #                 translated_value = ursgal_value
        #     translated_params[ mapped_unode_param ] = translated_value
        # return translated_params


    def time_point( self, tag=None, diff=True, format_time=False, stop=False ):
        '''
        Stores time_points in self.stats['time_points'] given a tag.
        returns time since tag was inserted if tag already exists.

        '''
        value = None
        if tag is not None:
            now = time.time()
            if tag in self.stats['time_points'].keys():
                start_time = self.stats['time_points'][ tag ]['start']

                if 'stop' in self.stats['time_points'][ tag ].keys():
                    now = self.stats['time_points'][ tag ]['stop']
                if diff:
                    value = now - start_time
                    if stop:
                        self.stats['time_points'][ tag ]['stop'] = now
                else:
                    value = start_time

            else:
                self.stats['time_points'][ tag ] = {'start' : now }
                value = now

            if format_time:
                value = time.ctime(
                    value
                )
        return value

    def update_history_status( self, engine=None, history=None, status='pending', history_addon=None ):
        if history is None:
            raise Exception("Legacy code implicitly updated history ... please change code!")

            # history = self.stats['history']
        if engine is None:
            engine = self.engine
        history = self.update_history_with_basic_unode_info(
            history= history,
            engine = engine
        )
        history[-1]['status'] = status
        if history_addon is not None:
            history[-1]['history_addon'] = history_addon
        return history

    def update_history_with_basic_unode_info(self, history=None, engine=None):
        # careful, self.engine != engine!

        #last istory entry
        init_empty_history_entry = False
        if len( history ) == 0:
            init_empty_history_entry = True
        else:
            last_history_entry = history[-1]

            if last_history_entry['engine'] != engine:
                init_empty_history_entry = True

            if last_history_entry['engine'] == engine and last_history_entry['status'] == 'complete':
                init_empty_history_entry = True

        if init_empty_history_entry:
            history.append(
                {
                    'engine'         : engine,
                    'ursgal_version' : ursgal.__version__,
                    'META_INFO'      : self.meta_unodes[ engine ].META_INFO,  # NOT self.META_INFO because self is the UController and not the search engine!!!
                    'finfo'          : self.io['input']['finfo']
                }
            )
        return history

        # entries_in_history = 0
        # last_engine_history_pos = 0
        # for pos, entry in enumerate(history):
        #     if entry['engine'] == engine:
        #         entries_in_history += 1
        #         last_engine_history_pos = pos

        # assert entries_in_history <= 1, '''
        # Found multiple entries in the history for this engine ({0}).
        # This should not happen. Consider re-running the pipeline.
        # '''.format( enigne )
        # if entries_in_history == 0:
        #     history.append(
        #         {
        #             'engine'    : engine,
        #             'startups'  : [],
        #             'META_INFO' : self.meta_unodes[ engine ].META_INFO,  # NOT self.META_INFO because self is the UController and not the search engine!!!
        #             'finfo'     : self.io['input']['finfo']
        #         }
        #     )
        # history_entry = history[-1]
        # assert history_entry['engine'] == engine, '''
        # Last entry in history is not from this engine ({0}).
        # This should not happen. Consider re-running the pipeline.

        # history : {1}
        # '''.format( engine, history  )
        # return history


if __name__ == '__main__':
    print('Yesss!')



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import time


def main():
    '''

    Testscript to evaluate the speed of common search engines.

    Usage:
        ./compare_search_engine_speed.py

    Note:
        The complete time for the search including peptide mapping and
        unifying the csv results.

    '''
    engine_list = [
        # 'msfragger_20170103',
        'xtandem_vengeance',
        'omssa',
        'msgfplus_v2016_09_16'
    ]

    params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'modifications': [
            '*,opt,Prot-N-term,Acetyl',
            'M,opt,any,Oxidation',
        ],
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ],
        'ftp_url': 'ftp.peptideatlas.org',
        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',
        'ftp_include_ext': [
            'JB_FASP_pH8_2-3_28122012.mzML',
        ],
        'ftp_output_folder': os.path.join(
            os.pardir,
            'example_data',
            'xtandem_and_msfragger_comparison'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        ),
    }

    if os.path.exists(params['ftp_output_folder']) is False:
        os.mkdir(params['ftp_output_folder'])

    uc = ursgal.UController(
        profile='LTQ XL low res',
        params=params
    )
    mzML_file = os.path.join(
        params['ftp_output_folder'],
        params['ftp_include_ext'][0]
    )
    if os.path.exists(mzML_file) is False:
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )
    time_collector = {}
    filtered_files_list = []
    mgf_file = uc.convert(
        input_file=mzML_file,
        engine='mzml2mgf_1_0_0',
    )
    for engine in engine_list:
        start_time = time.time()
        unified_result_file = uc.search(
            input_file=mzML_file,
            engine=engine,
        )
        time_collector[engine] = time.time() - start_time
        validated_file = uc.validate(
            input_file=unified_result_file,
            engine='percolator_2_08',
        )

        filtered_file = uc.execute_misc_engine(
            input_file=validated_file,
            engine='filter_csv_1_0_0',
        )

        filtered_files_list.append(filtered_file)

    uc.visualize(
        input_files=filtered_files_list,
        engine='venndiagram_1_1_0',
    )

    for key, duration in time_collector.items():
        print('Engine {0} took {1:1.2f}s'.format(key, duration))
    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import csv
import copy
import re
import pprint
import gzip
import subprocess
from .msgfplus2csv_v2017_07_04 import msgfplus2csv_v2017_07_04 as msgfc


class msgfplus2csv_v1_2_0( msgfc ):
    """
    msgfplus_C_mzid2csv_v1.2.0 UNode
    Parameter options at https://omics.pnl.gov/software/ms-gf

    Reference:
    Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, Mohammed S, Heck AJ, Pevzner PA. (2010) The Generating Function of CID, ETD, and CID/ETD Pairs of Tandem Mass Spectra: Applications to Database Search.
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'msgfplus_C_mzid2csv',
        'version'            : 'v1.2.0',
        'release_date'       : '2018-02-01',
        'engine_type' : {
            'converter'     : True
        },
        'input_extensions'   : ['.mzid', '.mzid.gz'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : None,
        'include_in_git'     : False,
        'in_development'     : False,
        'distributable'      : True,
        'utranslation_style' : 'msgfplus_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'     : 'MzidToTsvConverter.exe',
                    'url'     : '',
                    'zip_md5' : 'e6923e1f1726f0056d3e9b8bf31016ca'
                },
            },
        },
        'citation' : \
            'Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, '\
            'Mohammed S, Heck AJ, Pevzner PA. (2010) The Generating Function '\
            'of CID, ETD, and CID/ETD Pairs of Tandem Mass Spectra: '\
            'Applications to Database Search.',
    }



#!/usr/bin/env python3

from setuptools import setup
import setuptools
from setuptools.command.install_lib import install_lib
import setuptools.command.build_py
import distutils.cmd
import os
import sys


executable_list = [
    'makeblastdb',
    'tandem',
    'tandem.exe',
    'qvality',
    'percolator',
    'percolator_2_08',
    'omssacl',
    'myrimatch_2_1_138',
    'myrimatch.exe',
    'MSAmanda.exe',
    'PepNovo_bin',
    'novor.bat',
    'novor.sh',
    'PepNovo.exe',
]

if sys.platform in ['win32']:
    class my_install_lib(setuptools.command.install_lib.install_lib):
        pass
else:
    class my_install_lib(setuptools.command.install_lib.install_lib):

        def run(self):
            setuptools.command.install_lib.install_lib.run(self)
            for fn in self.get_outputs():
                if os.path.basename(fn) in executable_list:
                    # copied from setuptools source - make the binaries
                    # executable
                    mode = ((os.stat(fn).st_mode) | 0o555) & 0o7777
                    print("changing mode of %s to %o", fn, mode)
                    os.chmod(fn, mode)


class BuildPyWithResources(setuptools.command.build_py.build_py):
    """Includes install_resources.py before the setuptools build"""

    def run(self):
        self.run_command('install_resources')
        setuptools.command.build_py.build_py.run(self)


class InstallResourcesCommand(distutils.cmd.Command):
    """Download resources from webpage and install into ursgal/resources"""
    description = 'Download and install third party engines'
    user_options = []

    def initialize_options(self):
        return
    def finalize_options(self):
        return

    def run(self):
        '''
        Download all resources from our webpage to ursgal/resources.

        '''
        working_directory = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
        )
        if os.path.exists(os.path.join(working_directory, 'ursgal')) is False:
            print('Could not find ursgal directory')
            sys.exit(1)
        import ursgal
        uc = ursgal.UController()
        downloaded_zips = uc.download_resources(resources=None)
        if len(downloaded_zips) == 0:
            print('[ INFO ] No engines were downloaded, all should be available')
        else:
            print(
                '[ INFO ] Downloaded and installed {0} engine(s)'.format(
                    len(downloaded_zips)
                )
            )
            for engine, zip_file in downloaded_zips:
                print(
                    '[ INFO ] Engine: {0} has been installed from {1}'.format(
                        engine,
                        zip_file
                    )
                )

# We store our version number in a simple text file:
version_path = os.path.join(
    os.path.dirname(__file__),
    'ursgal', 'version.txt'
)
with open(version_path, 'r') as version_file:
    ursgal_version = version_file.read().strip()

with open('requirements.txt') as req:
    requirements = req.readlines()

setup(
    name='ursgal',
    version=ursgal_version,
    packages=['ursgal'],
    package_dir={'ursgal': 'ursgal'},
    description='ursgal',
    package_data={
        'ursgal' : [
            'version.txt',
            'wrappers/*.py',
            'resources/*/*/*',
            'resources/*/*/*/*',
            'resources/*/*/*/*/*',
        ]
    },
    build_requires=[
        'numpy',
    ],
    install_requires=requirements,
    long_description='Universal Python module combining common bottom-up proteomics tools for large-scale analysis',
    author='Lukas P. M. Kremer, Purevdulam Oyunchimeg, Johannes Barth, Stefan Schulze and Christian Fufezan',
    author_email='christian@fufezan.net',
    url='http://ursgal.github.com',
    license='Lesser GNU General Public License (LGPL)',
    platforms='any that supports python 3.5',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: SunOS/Solaris',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    cmdclass={
        'install_lib': my_install_lib,
        'build_py': BuildPyWithResources,
        'install_resources': InstallResourcesCommand,
    }
)



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Simple example script how to generate a target decoy database.

    Note:
        By default a 'shuffled peptide preserving cleavage sites' database is 
        generated. For this script a 'reverse protein' database is generated.

    usage:

        ./target_decoy_generation_example.py

    '''
    params = {
        'enzyme': 'trypsin',
        'decoy_generation_mode': 'reverse_protein',
    }

    fasta_database_list = [
        os.path.join(
            os.pardir,
            'example_data',
            'BSA.fasta'
        )
    ]

    uc = ursgal.UController(
        params=params
    )

    new_target_decoy_db_name = uc.execute_misc_engine(
        input_file=fasta_database_list,
        engine='generate_target_decoy_1_0_0',
        output_file_name='my_BSA_target_decoy.fasta',
    )
    print('Generated target decoy database: {0}'.format(
        new_target_decoy_db_name))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
import ursgal
import sys
import os
import pprint
import subprocess

class mzidentml_lib_1_6_10( ursgal.UNode ):
    """
    MzidLib 1_6_10 UNode

    'Reisinger F, Krishna R, Ghali F, Ríos D, Hermjakob H, Vizcaíno JA, Jones AR. (2012)
    jmzIdentML API: A Java interface to the mzIdentML standard for peptide and protein identification data.'

    Java program to convert results to .mzIdentML and .mzIdentML to .csv

    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'MzidLib',
        'version'            : '1.6.10',
        'release_date'       : None,
        'engine_type' : {
            'converter' : True
        },
        'input_extensions'   : ['.xml', '.xml.gz', '.csv', '.mzid', '.mzid.gz'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : None,
        'in_development'     : False,
        'include_in_git'     : False,
        'distributable'      : True,
        'utranslation_style' : 'mzidentml_lib_1_6_10',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'            : 'mzidentml-lib-1.6.10.jar',
                    'url'            : '',
                    'zip_md5'        : '849e00f50d814d3c44b29466434c2cfb',
                    'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Reisinger F, Krishna R, Ghali F, Rios D, Hermjakob H, '\
            'Vizcaino JA, Jones AR. (2012) jmzIdentML API: A Java interface '\
            'to the mzIdentML standard for peptide and protein identification '\
            'data.',
    }

    def __init__(self, *args, **kwargs):
        super(mzidentml_lib_1_6_10, self).__init__(*args, **kwargs)

    def raw2mzid( self, search_engine=None, translations=None ):
        '''
        Convert raw result files into .mzid result files
        '''
        if self.io['input']['finfo']['is_compressed']:
            self.params['translations']['mzidentml_compress'] = True
        else:
            self.params['translations']['mzidentml_compress'] = False

        if 'tandem' in search_engine:
            tmp_options = [
                '-databaseFileFormatID', 'MS:1001348',
                'massSpecFileFormatID', 'MS:1001062',
                '-idsStartAtZero', 'false'
            ]
            converter_mode = 'Tandem2mzid'
            translations['mzidentml_function']['mzidentml_function'] = converter_mode
            self.print_info('Executing xtandem xml to mzid conversion')
        else:
            self.print_info('''
    Do not know how to convert search results from {input_file} to mzid
            '''.format( **self.params ),caller='Warning')
            sys.exit(1)

        tmp_command_list = [
            'java',
            '-jar',
            self.exe,
        ]

        for translated_key, translation_dict in sorted(translations.items()):
                print(translated_key)
                if translated_key == '-Xmx':
                    self.params[ 'command_list' ].insert(1,'{0}{1}'.format(
                        translated_key,
                        list(translation_dict.values())[0]
                    ))
                elif translated_key == 'mzidentml_function':
                    self.params[ 'command_list' ].insert(4, list(translation_dict.values())[0])
                elif translated_key == 'output_file_incl_path':
                    self.params[ 'command_list' ].insert(6, list(translation_dict.values())[0])
                elif len(translation_dict) == 1:
                    self.params['command_list'].extend((translated_key, str(list(translation_dict.values())[0])))
                else:
                    print('The translatd key ', translated_key, ' maps on more than one ukey, but no special rules have been defined')
                    print(translation_dict)
                    sys.exit(1)

        tmp_command_list += tmp_options
        proc = subprocess.Popen(
            tmp_command_list,
            stdout=subprocess.PIPE,
        )
        for line in proc.stdout:
            print(line.strip().decode('utf'))
        return

    def preflight( self ):
        '''
        Convert .mzid result files from different search engines
        into .csv result files

        For X!Tandem result files first need to be converted into .mzid
        with raw2mzid
        '''

        translations = self.params['translations']['_grouped_by_translated_key']
        # import pprint
        # pprint.pprint(translations)
        # exit(1)

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        translations['output_file_incl_path']['output_file_incl_path'] = self.params['translations']['output_file_incl_path']

        search_engine = self.get_last_search_engine(
            history = self.stats['history']
        )
        if search_engine is None:
            raise TypeError(
                '\nmzidentml_lib UNode: Last search engine was not found in '\
                'history. Can\'t convert results without knowing which '\
                'search engine was used.'
            )

        FULL_INPUT_PATH = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        if 'tandem' in search_engine:
            # xtandem uses xml output which has first to be converted to .mzid
            self.raw2mzid( search_engine=search_engine, translations=translations)
            run_mz_ident_2_csv = True
        elif 'omssa' in search_engine:
            run_mz_ident_2_csv = False
        elif 'msgf' in search_engine:
            run_mz_ident_2_csv = True
        elif 'myrimatch' in search_engine:
            run_mz_ident_2_csv = True
        elif 'msamanda' in search_engine:
            run_mz_ident_2_csv = False
        else:
            raise Exception("Don't know the search engine that created your file...")

        if run_mz_ident_2_csv:
            assert os.path.exists(FULL_INPUT_PATH), '''
                No mzident file found {0}
            '''.format( FULL_INPUT_PATH )

            self.params[ 'command_list' ] = [
                'java',
                '-jar',
                self.exe,
                FULL_INPUT_PATH,
            ]

            for translated_key, translation_dict in sorted(translations.items()):
                if translated_key == '-Xmx':
                    self.params[ 'command_list' ].insert(1,'{0}{1}'.format(
                        translated_key,
                        list(translation_dict.values())[0]
                    ))
                elif translated_key == 'mzidentml_function':
                    self.params[ 'command_list' ].insert(4, list(translation_dict.values())[0])
                elif translated_key == 'output_file_incl_path':
                    self.params[ 'command_list' ].insert(6, list(translation_dict.values())[0])
                elif len(translation_dict) == 1:
                    self.params['command_list'].extend((translated_key, str(list(translation_dict.values())[0])))
                else:
                    print('The translatd key ', translated_key, ' maps on more than one ukey, but no special rules have been defined')
                    print(translation_dict)
                    exit(1)
            # print( ' '.join(self.params[ 'command_list' ]) )

        else:
            #OMSSA and MS amanda does not need a conversion
            self.params['command_list'] = []
        return



#!/usr/bin/env python3
import ursgal
import os


def main():
    """."""
    params = {
        'csv_filter_rules': [
            ['combined PEP', 'lte', 0.05],
            ['Is decoy', 'equals', 'false']
        ],
        'compress_raw_search_results_if_possible': False,
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'ftp_url': 'ftp.peptideatlas.org',
        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',
        'ftp_include_ext': [
            'JB_FASP_pH8_2-3_28122012.mzML',
        ],
        'ftp_output_folder': os.path.join(
            os.pardir,
            'example_data',
            'quantify_JB_example'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        )
    }
    if os.path.exists(params['ftp_output_folder']) is False:
        os.mkdir(params['ftp_output_folder'])

    engines = [
        'msgfplus_v2019_01_22',
        'xtandem_vengeance',
        # 'msfragger_20170103'
    ]

    uc = ursgal.UController(
        profile='LTQ XL low res',
        params=params
    )
    mzML_file = os.path.join(
        params['ftp_output_folder'],
        params['ftp_include_ext'][0]
    )

    if os.path.exists(mzML_file) is False:
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    mgf = uc.convert(
        input_file=mzML_file,
        engine='mzml2mgf_1_0_0'
    )

    all_res = []
    for label in ['14N', '15N']:
        uc.params['label'] = label
        uc.params['prefix'] = label
        results = []
        for engine in engines:
            res = uc.search(
                input_file=mgf,
                engine=engine
            )
            val = uc.validate(
                input_file=res,
                engine='percolator_2_08'
            )
            results.append(val)

        combined_results = uc.combine_search_results(
            input_files=results,
            engine='combine_PEP_1_0_0',
        )

        fil = uc.filter_csv(
            input_file=combined_results
        )
        all_res.append(fil)

    uc.params['quantification_evidences'] = all_res
    uc.params['label'] = '15N'
    uc.params['label_percentile'] = [0, 0.99]
    uc.params['evidence_score_field'] = 'combined PEP'
    uc.params['ms_level'] = 1

    uc.quantify(
        input_file=mzML_file,
        engine='pyqms_1_0_0',
        force=False
    )

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8

"""

Retrieve data via http protocol

usage:
    get_http_files_1_0_0.py <http_url>

"""

# import glob
from urllib import request as request
import os
import tempfile
import urllib
import string


def format_filename(s):
    """
    Take a string and return a valid filename constructed from the string.
    Any characters not present in valid_chars are removed.
    """
    valid_chars = "-_.()=# " + string.ascii_letters + string.digits
    filename = "".join(char for char in s if char in valid_chars)
    return filename


def main(http_url=None, http_output_folder=None):
    # retrieve files via http
    assert (
        http_url is not None
    ), "[ - HTTP - ] Require http_url not None to run ;)"

    print("[ - HTTP - ] Downloading files from {0} ...".format(http_url))

    basename = format_filename(os.path.basename(http_url))

    if http_output_folder is None:
        http_output_folder = tempfile.gettempdir()
    else:
        if os.path.exists(http_output_folder) is False:
            os.makedirs(http_output_folder)

    output_path = os.path.join(http_output_folder, basename)

    try:
        with open(output_path, "wb") as ooo:
            local_filename, headers = request.urlretrieve(
                http_url, filename=output_path
            )
        print("[ - HTTP - ] Saved file as {0}".format(output_path,))
    except urllib.error.URLError:
        print(
            "[ - HTTP - ] \t WARNING! Could not download {0} Check your internet connection!".format(
                http_url
            ),
            "[ - HTTP - ] For OSX, make sure that certificates are installed (/Applications/Python 3.x/Install Certificates.command)",
        )
        os.remove(output_path)
    return output_path


if __name__ == "__main__":
    main(
        http_url="https://cdn.vox-cdn.com/thumbor/SMiHkpdBdVzg-XRkx_kmQSRdRp4=/0x0:1579x816/920x613/filters:focal(603x200:909x506)/cdn.vox-cdn.com/uploads/chorus_image/image/61585025/pulp_fiction_brett_kavanaugh.0.png",
        http_output_folder=".",
    )



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import csv
import copy
import re
import pprint
from .msgfplus2csv_v2016_09_16 import msgfplus2csv_v2016_09_16 as msgf2csv


class msgfplus2csv_v2017_01_27( msgf2csv ):
    """
    msgfplus2csv_v2017_01_27 UNode
    Parameter options at https://omics.pnl.gov/software/ms-gf

    Reference:

        Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, Mohammed S,
        Heck AJ, Pevzner PA. (2010) The Generating Function of CID, ETD, and
        CID/ETD Pairs  of Tandem Mass Spectra: Applications to Database Search.
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'msgfplus2csv',
        'version'            : 'v2017.01.27',
        'release_date'       : '2017-1-27',
        'engine_type' : {
            'converter'     : True
        },
        'input_extensions'   : ['.mzid', '.mzid.gz'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : None,
        'include_in_git'     : False,
        'in_development'     : False,
        'distributable'  : False,
        'utranslation_style' : 'msgfplus_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'     : 'MSGFPlus.jar',
                    'url'     : '',
                    'zip_md5' : '',
                },
            },
        },
        'citation' : \
            'Kim S, Mischerikow N, Bandeira N, Navarro JD, Wich L, '\
            'Mohammed S, Heck AJ, Pevzner PA. (2010) The Generating Function '\
            'of CID, ETD, and CID/ETD Pairs of Tandem Mass Spectra: '\
            'Applications to Database Search.',
        'uses_unode' : 'msgfplus_v2017_01_27',
    }



#!/usr/bin/python

import csv
import functools
import operator
from collections import defaultdict


def naive_bayes(prob_list):
    '''
    Independent probabilities a, b, c can be combined like this:

                             a*b*c
    combined_prob = -------------------------
                    a*b*c + (1-a)*(1-b)*(1-c)

    For a straightforward explanation, see
    http://www.paulgraham.com/naivebayes.html
    '''

    # multiplying all probabilities: a*b*c
    multiplied_probs = \
        functools.reduce(operator.mul, prob_list, 1)
    # multiplying the opposite of all probabilities: (1-a)*(1-b)*(1-c)
    multiplied_opposite_probs = \
        functools.reduce(operator.mul, (1-p for p in prob_list), 1)
    return multiplied_probs / (multiplied_probs + multiplied_opposite_probs)


def get_row_key(row, columns_for_grouping):
    return tuple( row[c] for c in columns_for_grouping )
        

def list_to_sorted_tuple(l):
    return tuple(sorted(set(l)))


def main(columns_for_grouping=None, input_csvs=None, output_csv=None,
         input_sep=None, output_sep=None, join_sep=None, pep_colname=None,
         input_engines=None):
    '''
    Merges CSVs from different search engines and adds a column called
    'combined PEP' that contains the combination of all engine PEPs
    as computed with the naive Bayes method.
    '''
    assert len(input_csvs) > 1, 'Input must be at least 2 files.'
    columns_for_grouping = list_to_sorted_tuple(columns_for_grouping)

    merged_fieldnames_set = set(['combined PEP', 'engines'])
    merged_dict = defaultdict(dict)
    for input_csv, input_engine in zip(input_csvs, input_engines):
        # give the engine its own PEP column:
        engine_PEP_colname = ':'.join([input_engine, pep_colname])
        merged_fieldnames_set.add(engine_PEP_colname)
        with open(input_csv, 'r', encoding='utf8') as csv_object:
            reader = csv.DictReader(csv_object, delimiter=input_sep)
            merged_fieldnames_set |= set(reader.fieldnames)
            for row in reader:
                row['engines'] = input_engine
                row_key = get_row_key(row, columns_for_grouping)
                for col, val in row.items():
                    if col == pep_colname:
                        merged_dict[row_key][engine_PEP_colname] = val
                        # store the PEP of that PSM in a list that is
                        # used to compute combined PEP later
                        if 'combined PEP' not in merged_dict[row_key]:
                            merged_dict[row_key]['combined PEP'] = []
                        merged_dict[row_key]['combined PEP'].append(val)
                        continue
                    if col not in merged_dict[row_key]:
                        merged_dict[row_key][col] = val
                    elif val == merged_dict[row_key][col]:
                        continue
                        # value is already there, no need to do anything
                    elif val != merged_dict[row_key][col]:
                        #print("WARNING! Conflicting field values in column {0}: {1} and {2}".format(col, val, merged_dict[row_key][col]))
                        old_vals = set(merged_dict[row_key][col].split(join_sep))
                        old_vals.add(val)
                        old_vals.discard('')
                        merged_dict[row_key][col] = join_sep.join(sorted(old_vals))
                    else:
                        raise Exception('This is impossibru!')


    merged_fieldnames = list(columns_for_grouping) + sorted(merged_fieldnames_set - set(columns_for_grouping))
    merged_fieldnames.remove(pep_colname)
    with open(output_csv, 'w', encoding='utf8') as outf:
        writer = csv.DictWriter(outf, fieldnames=merged_fieldnames, delimiter=output_sep)
        writer.writeheader()
        for out_row in merged_dict.values():
            all_PEPs = [float(p) for p in out_row['combined PEP']]
            out_row['combined PEP'] = naive_bayes(all_PEPs)
            writer.writerow(out_row)
    return


def parse_args(verbose=True):
    '''
    Parses command line arguments and returns them in a dict.
    Only used when executing this script without Ursgal.
    '''
    import argparse
    # no need to import argparse when script is executed from ursgal

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-i', '--input_csvs', type=str, nargs='+', required=True,
        help='Paths to unified input CSV files (2 or more)')
    parser.add_argument(
        '-c', '--columns_for_grouping', type=str, nargs='+', required=True,
        help='Column names by which the rows should be grouped')
    parser.add_argument(
        '-o', '--output_csv', type=str, help='Output CSV name')
    parser.add_argument(
        '-is', '--input_sep', type=str, default=',',
        help='Input file column delimiter character')
    parser.add_argument(
        '-os', '--output_sep', type=str, default=',',
        help='Output file column delimiter character')
    parser.add_argument(
        '-js', '--join_sep', type=str, default=';',
        help='Delimiter for multiple values in the same field')
    parser.add_argument(
        '-e', '--input_engines', type=str, required=True,
        help='The search engines of each input file (must be same order as input_csvs)')

    args = vars(parser.parse_args())  # convert to dict
    if verbose:
        print('You are using the following settings:')
        for arg, val in sorted(args.items()):
            print('  {0: <13}:  {1}'.format(arg, val))
        print()
    return args


if __name__ == '__main__':
    command_line_args = parse_args()
    main(**command_line_args)


#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import glob
import os.path
import sys
import tempfile
import time


def main():
    '''
    Testscript for executing the test node, which also tests the run time
    determination function.

    Usage:
        ./test_node_excution.py

    '''
    uc = ursgal.UController(verbose=True)
    temp_int, temp_fpath = tempfile.mkstemp(
        prefix='ursgal_',
        suffix='.csv'
    )
    temp_fobject = open(temp_fpath, 'w')
    print('test 1,2,3', file=temp_fobject)
    temp_fobject.close()
    test_1 = uc.execute_unode(
        temp_fpath,
        '_test_node'
    )
    test_2 = uc.execute_unode(
        test_1,
        '_test_node'
    )

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import sys
import glob


def main(input_path=None):
    '''
    Convert a .raw file to .mzML using the ThermoRawFileParser.
    The given argument can be either a single file or a folder 
    containing raw files.

    Usage:
        ./convert_raw_to_mzml.py <raw_file/raw_file_folder>
    '''
    R = ursgal.UController()

    # Check if single file or folder.
    # Collect raw files if folder is given
    input_file_list = []
    if input_path.lower().endswith('.raw'):
        input_file_list.append(input_path)
    else:
        for raw in glob.glob(os.path.join('{0}'.format(input_path), '*.raw')):
            input_file_list.append(raw)

    # Convert raw file(s)
    for raw_file in input_file_list:
        mzml_file = R.convert(
            input_file=raw_file,
            engine='thermo_raw_file_parser_1_1_2',
        )

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(main.__doc__)
    main(input_path=sys.argv[1])



#!/usr/bin/env python
# encoding: utf-8
"""
    Ursgal MappOrs

    :copyright: (c) 2014 by C. Fufezan, S. Schulze
    :licence: BSD, see LISCENSE for more details

"""
import ursgal
from collections import defaultdict as ddict
import multiprocessing
import os
import time

try:
    import regex as regex
    finditer_kwargs = { 'overlapped' : True }
except:
    print('[ WARNING  ] Standard re module cannot find overlapping pattern')
    print('[   INFO   ] Consider installing the regex module')
    print('[   INFO   ] pip install -r requirements.txt')
    import re as regex
    finditer_kwargs = {}


class UParamMapper( dict ):
    '''
    UParamMapper class offers interface to ursgal.uparams

    By default, the ursgal.uparams are parsed and the UParamMapper class
    is set with the ursgal_params dictionary.
    '''
    version = '0.2.0'

    def __init__( self, *args, **kwargs):
        # kwargs['engine_path'] = ursgal.__file__
        super(UParamMapper, self).__init__(*args, **kwargs)

        assert len(args) <= 1, 'Can only be initialized with max one argument'
        if len(args) == 0:
            params_dict = ursgal.uparams.ursgal_params
            # urgsal_dict
        else:
            params_dict = args[0]

        self.update( params_dict )
        self.lookup = self.group_styles()
        self._eval_functions = {
            'cpus' : {
                'max'     : multiprocessing.cpu_count(),
                'max - 1' : multiprocessing.cpu_count() - 1,
                'max-1'   : multiprocessing.cpu_count() - 1,
            }
        }

    def _assert_engine(self, engine):
        assert engine is not None, 'must define engine!'
        return

    def _eval_default_value( self, ukey, uvalue ):
        rvalue = None
        if ukey in self._eval_functions.keys():
            if uvalue in self._eval_functions[ ukey ].keys():
                rvalue = self._eval_functions[ ukey ][ uvalue ]
                if ukey == 'cpus' and rvalue == 0:
                    rvalue = 1

        return rvalue

    def mapping_dicts( self, engine_or_engine_style):
        '''yields all mapping dicts'''
        if '_style_' in engine_or_engine_style:
            lookup_key = 'style_2_params'
            style = engine_or_engine_style
        else:
            lookup_key = 'engine_2_params'
            style = self.lookup['engine_2_style'].get(engine_or_engine_style, None)

        for uparam in self.lookup[ lookup_key ].get(engine_or_engine_style, []):
            sup = self[ uparam ]

            uvalue_style_translation = sup['uvalue_translation'].get(style, {})
            assert isinstance(uvalue_style_translation, dict), '''
            Syntax error in ursgal/uparams.py at key {0}
            '''.format( uparam )
            if len(uvalue_style_translation.keys()) != 0:
                # can be translated, at least by some engine
                # and thus is not a list of elements ;)
                translated_value = uvalue_style_translation.get(
                    sup['default_value'],
                    sup['default_value']
                )
            else:
                translated_value = sup['default_value']

            if '_uevaluation_req' in sup.get('uvalue_type',''):
                re_evaluated_value = self._eval_default_value(
                    uparam,
                    translated_value
                )
                if re_evaluated_value is not None:
                    translated_value = re_evaluated_value

            template = sup.copy()
            keys_to_delete = [
                'ukey_translation',
                'uvalue_translation',
                'available_in_unode'
            ]
            for k in keys_to_delete:
                del template[ k ]
            template.update(
                {
                    'style'                    : style,
                    'ukey'                     : uparam,
                    'ukey_translated'          : sup['ukey_translation'][ style ],
                    'default_value_translated' : translated_value,
                    'uvalue_style_translation' : uvalue_style_translation,
                    'triggers_rerun'           : sup.get('triggers_rerun', True)
                }
            )

            yield template

    def get_masked_params( self, mask = None):
        '''
        Lists all uparams and the fields specified in the mask

        For example::

            upapa.get_masked_params( mask = ['uvalue_type']) will return::
                {
                    '-xmx' : {
                        'uvalue_type' : "str",
                    },
                    'aa_exception_dict' : {
                        'uvalue_type' : "dict",
                    },
                    ...
                }

        '''
        if mask is None:
            mask = []
        masked_params = {}
        for key, value in self.items():
            masked_params[ key ] = {}
            for mkey in mask:
                masked_params[ key ][ mkey ] = value.get(mkey, None)
        return masked_params


    def get_all_params( self, engine=None):
        self._assert_engine( engine )
        all_params = []
        for ukey in self.keys():
            if engine in self[ ukey ]['available_in_unode']:
                all_params.append( ukey )
        return all_params

    def group_styles(self):
        '''
        Parses self.items() and build up lookups.
        Additionally, consistency check is performed to guarantee that each
        engine is mapping only on one style.

        The lookup build and returned looks like::

            lookup = {
                'style_2_engine' : {
                    'xtandem_style_1' : [
                        'xtandem_sledgehamer',
                        'xtandem_cylone',
                        ...
                    ],
                    'omssa_style_1' ...
                },
                # This is done during uNode initializations
                # each unode will register its style with umapmaster
                #
                'engine_2_style' : {
                    'xtandem_sledgehamer' : 'xtandem_style_1', ...
                },
                'engine_2_params' : {
                    'xtandem_sledgehamer' : [ uparam1, uparam2, ...], ...
                },
                'style_2_params' : {
                    'xtandem_style_1' : [ uparam1, uparam2, ... ], ...
                },
                'params_triggering_rerun' : {
                    'xtandem_style_1' : [ uparam1, uparam2 .... ]
                }
            }
        '''
        lookup = {
            'style_2_engine' : ddict(set),
            'engine_2_style' : {},
            # these two are not in the docu yet ...
            'engine_2_params': ddict(list),
            'style_2_params': ddict(list),
            'params_triggering_rerun' : ddict(list)
        }
        for uparam, udict in sorted(self.items()):
            # print( uparam, end = '\t')
            # if uparam == 'force':
            #     print(udict)
            for style in udict['ukey_translation'].keys():
                try:
                    style_basename, style_version = style.split('_style_')
                except:
                    print('Syntax Error @ uparam {0}'.format(uparam))
                    print('style : {0}'.format( style ))
                    exit(1)
                vvv = False
                if style_basename == 'ucontroller':
                    vvv = True
                    # print('UController params {0}'.format( uparam ))
                # else:
                    # print()
                styles_seen = set()
                for engine in udict['available_in_unode']:
                    # if vvv:
                    #     print(engine )
                    if style_basename not in engine:
                        continue
                    # this style 2 engine lookup is not quite right ...
                    # This function requires unode meta info for proper
                    # mapping ....
                    # lookup['style_2_engine'][ style ].add( engine )
                    lookup['engine_2_params'][ engine ].append( uparam )

                    if style not in styles_seen:
                        lookup['style_2_params'][ style ].append( uparam )
                        if udict.get('triggers_rerun', True):
                            lookup['params_triggering_rerun'][ style ].append( uparam )
                        styles_seen.add( style )

                    parsed_e2s = lookup['engine_2_style'].get( engine, None)
                    if parsed_e2s is None:
                        lookup['engine_2_style'][ engine ] = style
                    else:
                        if parsed_e2s != style:
                            print(
                                '{0} was found to map on style {1} and {2}'.format(
                                    engine, parsed_e2s, style
                                )
                            )
                            sys.exit(1)
        return lookup

    def _show_params_overview( self, engine=None):
        self._assert_engine( engine )
        # This can be done differently with the lookups now ...
        for param in sorted(self.get_all_params( engine=engine)):
            udefault_value     = self[ param ]['default_value']

            ukey_translation   = self[ param ]['ukey_translation'].get(
                'msgfplus_style_1', '??'
            )
            uvalue_translation = self[ param ]['uvalue_translation'].get(
                'msgfplus_style_1', {}
            )
            try:
                translated_default_param_value = uvalue_translation.get( param, 'n/d')
            except:
                translated_default_param_value = 'complex type'
            try:

                print(
                    'uParams {0: >30} : {1: <20} {2: >30} : {3: <20}'.format(
                        param,
                        udefault_value,
                        ukey_translation,
                        translated_default_param_value,
                    )
                )
            except:

                print('Failed on: ' , param)


if __name__ == '__main__':
    print('Yes!')



#!/usr/bin/env python3
# encoding: utf-8
import ursgal
R = ursgal.UController()

TESTS = [
    {
        'input' : {
            'user_fname'   : 'my_file.svg',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.svg'
    },

    {
        'input' : {
            'user_fname'   : 'my_file',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.svg'
    },

    {
        'input' : {
            'user_fname'   : 'my_file.png',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.png.svg'
    },

    {
        'input' : {
            'user_fname'   : 'my_file.png.gz',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.png.gz.svg'
    },

    {
        'input' : {
            'user_fname'   : '/my_directory/my_file.png.gz',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.png.gz.svg'
    },

    {
        'input' : {
            'user_fname'   : '/my_directory/my_file.png.gz',
            'engine'       : 'venndiagram_1_0_0',
        },
        'output' : 'my_file.png.gz.svg'
    },

    {
        'input' : {
            'user_fname'   : '/my_directory/my_file.svg',
            'engine'       : 'venndiagram_1_0_0',
        },
        'prefix' : 'xx',
        'output' : 'xx_my_file.svg'
    },

    {
        'input' : {
            'user_fname'   : '/my_directory/my_file.svg',
            'engine'       : 'venndiagram_1_0_0',
        },
        'prefix' : '',
        'output' : 'my_file.svg'
    },

    {
        'input' : {
            'user_fname'   : '/my_directory/my_file.svg',
            'engine'       : 'venndiagram_1_0_0',
        },
        'prefix' : None,
        'output' : 'my_file.svg'
    },
]


def sanitize_userdefined_output_filename_test():
    for test_id, test_dict in enumerate(TESTS):
        yield sanitize_userdefined_output_filename, test_dict


def sanitize_userdefined_output_filename( test_dict ):
    if 'prefix' in test_dict:
        R.params['prefix'] = test_dict['prefix']
    R.sanitize_userdefined_output_filename( **test_dict['input'] )
    out_put = R.sanitize_userdefined_output_filename(
        **test_dict['input']
    )
    print('Results', out_put , test_dict)
    assert out_put == test_dict['output'], '''
    sanitize_userdefined_output_filename {0} failed'''.format(
        test_dict
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        sanitize_userdefined_output_filename( test_dict )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import ursgal.unimod_mapper

M = ursgal.unimod_mapper.UnimodMapper()

UNIMODMAPPER_FUNCTIONS = [
    M.name2mass,
    M.name2mass,
    M.name2composition,
    M.name2id,
    M.id2mass,
    M.id2composition,
    M.id2name,
    M.mass2name_list,
    M.mass2id_list,
    M.mass2composition_list,
    M.appMass2id_list,
    M.appMass2element_list,
    M.appMass2name_list,
    M.composition2name_list,
    M.composition2id_list
]

TESTS = [
    [
        {
            'in' : {
                'args': ['ICAT-G:2H(8)']
            },
            'out': 494.30142
        },  # First test : ursgal.UnimodMapper.name2mass,
        {
            'in' : {
                'args': [' ICAT-G:2H(8)']
            },
            'out': 494.30142
        },  # second test that passes to many whitespaces :),
        {
            'in' : {
                'args': ['ICAT-G:2H(8)']
            },
            'out': {'H': 30, 'C': 22, 'O': 6, 'S': 1, '2H': 8, 'N': 4}
        },  # ursgal.UnimodMapper.name2composition,
        {
            'in' : {
                'args': ['ICAT-G:2H(8)']
            },
            'out' : '9'
        },  # ursgal.UnimodMapper.name2id,
        {
            'in' : {
                'args': ['9']
            },
            'out': 494.30142
        },  # ursgal.UnimodMapper.id2mass,
        {
            'in' : {
                'args': ['9']
            },
            'out' : {'N': 4, 'S': 1, '2H': 8, 'O': 6, 'C': 22, 'H': 30}
        },  # ursgal.UnimodMapper.id2composition,
        {
            'in' : {
                'args': '9'
            },
            'out' : 'ICAT-G:2H(8)'
        },  # ursgal.UnimodMapper.id2name,
        {
            'in' : {
                'args': [494.30142]
            },
            'out': ['ICAT-G:2H(8)']
        },  # ursgal.UnimodMapper.mass2name_list,
        {
            'in' : {
                'args': [494.30142]
            },
            'out' : ['9']
        },  # ursgal.UnimodMapper.mass2id_list,
        {
            'in' : {
                'args': [494.30142]
            },
            'out' : [{'N': 4, 'S': 1, '2H': 8, 'O': 6, 'C': 22, 'H': 30}]
        },  # ursgal.UnimodMapper.mass2composition_list,
        {
            'in' : {
                'args': [18],
                'kwargs': {'decimal_places': 0 }
            },
            'out': ['127', '329', '608', '1079', '1167']
        },  # ursgal.UnimodMapper.appMass2id_list
        {
            'in' : {
                'args': [18],
                'kwargs': {'decimal_places': 0 }
            },
            'out': [{'F': 1, 'H': -1}, {'13C': 1, 'H': -1, '2H': 3}, {'H': -2, 'C': -1, 'S': 1}, {'H': 2, 'C': 4, 'O': -2}, {'H': -2, 'C': -1, 'O': 2}]
        },  # ursgal.UnimodMapper.appMass2element_list
        {
            'in' : {
                'args': [18],
                'kwargs': {'decimal_places': 0 }
            },
            'out': ['Fluoro', 'Methyl:2H(3)13C(1)', 'Xle->Met', 'Glu->Phe', 'Pro->Asp']
        },  # ursgal.UnimodMapper.appMass2name_list
        {
            'in' : {
                'args': ['C(2)H(3)N(1)O(1)'],
            },
            'out': ['Carbamidomethyl', 'Ala->Gln', 'Gly->Asn', 'Gly']
        },
        {
            'in' : {
                'args': [ 'C(22)H(30)2H(8)N(4)O(6)S(1)'],
            },
            'out': ['9']
        },



        # end of first data set ...
    ]
]


def test_set_integrity_test():
    for test_id, list_of_test in enumerate(TESTS):
        yield input_list_check, list_of_test


def input_list_check( list_of_test ):
    '''
    Checks that the number of tests can actually be zipped
    with the unimodMapper functions
    '''
    assert len(list_of_test) <= len(UNIMODMAPPER_FUNCTIONS)


def unimodMapper_conversion_test():
    for test_id, list_of_test in enumerate(TESTS):
        test_function_association = zip( list_of_test, UNIMODMAPPER_FUNCTIONS )
        for test_dict, mapper_function in test_function_association:
            if 'kwargs' not in test_dict['in'].keys():
                test_dict['in']['kwargs'] = {}
            yield mapper_function_check, test_dict, mapper_function


def mapper_function_check( test_dict, mapper_function):
        mapper_output = mapper_function(
            *test_dict['in']['args'],
            **test_dict['in']['kwargs']
        )
        print(mapper_output,test_dict['out'])
        assert mapper_output == test_dict['out']


if __name__ == '__main__':
    print('Yes!')



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle


class upeptide_mapper_1_0_0( ursgal.UNode ):
    """
    upeptide_mapper_1_0_0 UNode

    Note:
        Different converter versions can be used (see parameter
        'peptide_mapper_converter_version') as well as different classes
        inside the converter node (see parameter
        'peptide_mapper_class_version' )

    Available converter classes of upeptide_mapper_1_0_0
        * UPeptideMapper_v3 (default)
        * UPeptideMapper_v4 (no buffering and enhanced speed to v3)
        * UPeptideMapper_v2
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'UpeptideMapper',
        'version'            : '1.0.0',
        'release_date'       : None,
        'engine_type' : {
            'misc_engine'     : True
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'output_suffix'     : 'pmap',
        'include_in_git'    : True,
        'in_development'    : False,
        'distributable'      : True,
        'utranslation_style': 'upeptide_mapper_style_1',
        'engine': {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'upeptide_mapper_1_0_0.py',
                },
            },
        },
        'citation'          : 'Kremer, L. P. M., Leufken, J., '\
            'Oyunchimeg, P., Schulze, S. & Fufezan, C. (2016) '\
            'Ursgal, Universal Python Module Combining Common Bottom-Up '\
            'Proteomics Tools for Large-Scale Analysis. '\
            'J. Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(upeptide_mapper_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        '''
        Peptides from search engine csv file are mapped to the given database(s)

        '''
        print('[ -ENGINE- ] Executing conversion ..')
        upeptide_mapper_main = self.import_engine_as_python_function()
        if self.params['output_file'].lower().endswith('.csv') is False:
            raise ValueError('Trying to use a non-csv file')

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        input_file  = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        tmp_files = upeptide_mapper_main(
            input_file      = input_file,
            output_file     = output_file,
            params          = self.params,
        )
        for tmp_file in tmp_files:
            self.created_tmp_files.append(tmp_file)

        return output_file



#!/usr/bin/env python3.2
# encoding: utf-8

import ursgal

TESTS = [
    {
        'input' : 'A#Oxidation:1',
        'output' : 'C(3)H(7)N(1)O(3)',
        'mod_pos_info' : [
            {
                'pos' : 1,
                'cc_mods': {'O': 1 }
            }
        ]
    },
    {
        'input' : 'A#Oxidation:1;Acetyl:0',
        # zero should end up at pos 1 ...
        'output' : 'C(5)H(9)N(1)O(4)',
        'mod_pos_info' : [
            {
                'pos' : 1,
                'cc_mods' : {'O': 2, 'C' : 2, 'H' : 2}
            }
        ]
    },
    {
        'input' : 'X',
        'aa_compositions' : {'X': 'C12'},
        'output' : 'C(12)H(2)O(1)',
        'mod_pos_info' : [
            {
                'pos' : 1,
                'cc_mods' : None
            }
        ]
    },
    {
        'input' : 'RAA',
        # 'aa_compositions' :  {'X': 'C12', 'R0': 'C12'},
        'output' : 'C(12)H(24)N(6)O(4)'
    },
    {
        'input' : 'R0X',
        'aa_compositions' : {'X': 'C12', 'R0': 'C12'},
        'output' : 'C(24)H(2)O(1)',
        # 'mod_pos_info' : (0, [{'C': 12 }], 1, {'O' : 1, 'C' : 2, 'H' : 2}])
    },
    {
        'input': 'R#Label:13C(6)15N(2):1',
        'output' : 'H(14)13C(6)15N(2)N(2)O(2)'
    },
    {
        'input' : 'R0XR1',
        'aa_compositions' : {'R0': 'C12', 'X': 'C13', 'R1': 'C14'},
        'output' : 'C(39)H(2)O(1)',
        'aa_pos_info' : [
            { 'pos' : 1, 'cc_mods' : {'C': 12 } },
            { 'pos' : 3, 'cc_mods' : {'C': 14 } }
        ],
    },
    {
        'input': 'RR#Label:13C(6)15N(2):1',
        'output' : 'C(6)H(26)13C(6)15N(2)N(6)O(3)',
        'aa_pos_info' : [
            { 'pos' : 1, 'cc_mods' : {'C': 6, 'O': 1, 'N': 4, 'H': 12} },
            { 'pos' : 2, 'cc_mods' : {'C': 6, 'O': 1, 'N': 4, 'H': 12} },
        ],
        'mod_pos_info' : [
            { 'pos' : 1, 'cc_mods': {'C': -6, '13C': 6, 'N': -2, '15N': 2}}
        ]
    },
    {
        'input': '+H2O2H2',
        'output' : 'H(4)O(2)',
    },
    {
        'input': '+H2O2H2-HO',
        'output' : 'H(3)O(1)',
    },
]


SUBTRACTION_TEST_SET = [
    # functionality tested yet ...
    {
        'in': '+H2O2H2',
        'sub' : 'H2',
        'out': {'H':2, 'O': 2},
    }
]

def pepitde_with_unimod_test():
    for test_id, test_dict in enumerate(TESTS):
        yield check_unimod_pep_input, test_dict


def check_unimod_pep_input( test_dict ):
    if 'aa_compositions' not in test_dict.keys():
        test_dict['aa_compositions'] = ursgal.chemical_composition_kb.aa_compositions
    cc = ursgal.ChemicalComposition(
        test_dict['input'],
        aa_compositions = test_dict['aa_compositions']
    )
    print('hill:', cc.hill_notation_unimod( ))
    print(cc.composition_of_aa_at_pos)
    print(cc.composition_of_mod_at_pos)
    assert cc.hill_notation_unimod( ) == test_dict['output']
    if 'mod_pos_info' in test_dict.keys():
        for mod_dict in test_dict['mod_pos_info']:
            cc_mods = cc.composition_of_mod_at_pos.get(mod_dict['pos'], None)
            assert cc_mods == mod_dict['cc_mods']
    if 'aa_pos_info' in test_dict.keys():
        for aa_dict in test_dict['aa_pos_info']:
            cc_aa = cc.composition_of_aa_at_pos.get( aa_dict['pos'], None )
            assert cc_aa == aa_dict['cc_mods']


def Simple_amino_acid_test():
    for aa, chemformula in ursgal.chemical_composition_kb.aa_compositions.items():
        yield check_hill_notation, aa, chemformula


def check_hill_notation( aa, chemformula ):
    cc = ursgal.ChemicalComposition( aa )
    cc.subtract_chemical_formula('H2O')
    print( aa, chemformula )
    print( cc )
    assert cc.hill_notation( ) == chemformula


if __name__ == '__main__':
    Simple_amino_acid_test()
